#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Config.py

A simple container class for the configuration of PyTyle. See pytylerc
for more information on configuring PyTyle.

The only variable here that isn't in pytylerc is Config.TILERS. This
configuration variable is loaded at program start up as a dict of all
available tiling algorithms (as Python modules).

There is also a getter method here, just in case we're looking for an
option that isn't specified in the config.
"""

class Config:
    #------------------------------------------------------------------------------
    # CONFIGURATION VARIABLES
    #------------------------------------------------------------------------------

    MISC = {}
    KEYMAP = {}
    WORKAREA = {}
    FILTER = []
    LAYOUT = {}
    TILING = {}
    TILERS = {}
    CALLBACKS = {}


    #------------------------------------------------------------------------------
    # CONFIGURATION DEFAULTS
    #------------------------------------------------------------------------------

    DEFAULTS = {
                'MISC': {
                         'tilers': ['Vertical', 'Horizontal', 'Maximal', 'Cascade'],
                         'global_tiling': False,
                         'timeout': 0.1,
                         'decorations': True,
                         'original_decor': True,
                         },
                'KEYMAP': {
                           'Alt-A': 'tile.default',
                           'Alt-U': 'untile',
                           'Alt-Z': 'cycle_tiler',
                           'Alt-Shift-space': 'reset',
                           'Alt-C': 'cycle',
                           'Alt-W': 'screen0_focus',
                           'Alt-E': 'screen1_focus',
                           'Alt-R': 'screen2_focus',
                           'Alt-Shift-W': 'screen0_put',
                           'Alt-Shift-E': 'screen1_put',
                           'Alt-Shift-R': 'screen2_put',
                           'Alt-H': 'master_decrease',
                           'Alt-L': 'master_increase',
                           'Alt-period': 'add_master',
                           'Alt-comma': 'remove_master',
                           'Alt-Return': 'make_active_master',
                           'Alt-M': 'win_master',
                           'Alt-Shift-C': 'win_close',
                           'Alt-J': 'win_previous',
                           'Alt-K': 'win_next',
                           'Alt-Shift-J': 'switch_previous',
                           'Alt-Shift-K': 'switch_next',
                           'Alt-X': 'max_all',
                           'Alt-S': 'restore_all',
                           },
                'WORKAREA': {
                             0: {
                                 'top': 0,
                                 'bottom': 0,
                                 'right': 0,
                                 'left': 0,
                                 },
                             },
                'FILTER': [
                           'gmrun', 'gimp', 'download',
                           ],
                'LAYOUT': {
                           'Vertical': {
                                        'width_factor': 0.5,
                                        'margin': 0,
                                        },
                           'Horizontal': {
                                          'height_factor': 0.5,
                                          'margin': 0,
                                          },
                           'Maximal': {},
                           'Cascade': {
                                       'decoration_height': 25,
                                       'width_factor': 1.0,
                                       'height_factor': 1.0,
                                       'push_over': 0,
                                       'horz_align': 'left',
                                       },
                           'HorizontalRows': {
                                              'row_size': 2,
                                              'height_factor': 0.5,
                                              'margin': 0,
                                              },
                           },
                'TILING': {
                           'default': 'Vertical',
                           },
                'CALLBACKS': {
                            0:'make_active_master',
                            1:'switch_previous',
                            2:'switch_next'
                            }
                }


    #------------------------------------------------------------------------------
    # CONFIGURATION GETTERS
    #------------------------------------------------------------------------------

    @staticmethod
    def misc(name):
        if name in Config.MISC:
            return Config.MISC[name]

        if name in Config.DEFAULTS['MISC']:
            return Config.DEFAULTS['MISC'][name]

        return None

    @staticmethod
    def keymap(keys):
        if keys in Config.KEYMAP:
            return Config.KEYMAP[keys]

        if keys in Config.DEFAULTS['KEYMAP']:
            return Config.DEFAULTS['KEYMAP'][keys]

        return None

    @staticmethod
    def workarea(screen_number, section):
        if screen_number in Config.WORKAREA and section in Config.WORKAREA[screen_number]:
            return Config.WORKAREA[screen_number][section]

        return Config.DEFAULTS['WORKAREA'][0][section]

    @staticmethod
    def filter():
        return Config.FILTER

    @staticmethod
    def layout(tiler, option):
        layout = tiler.__class__.__name__
        if layout in Config.LAYOUT and option in Config.LAYOUT[layout]:
            return Config.LAYOUT[layout][option]

        if layout in Config.DEFAULTS['LAYOUT'] and option in Config.DEFAULTS['LAYOUT'][layout]:
            return Config.DEFAULTS['LAYOUT'][layout][option]

        return None

    @staticmethod
    def tiling(screen, desk_or_view):
        if screen in Config.TILING:
            if isinstance(Config.TILING[screen], dict) and desk_or_view in Config.TILING[screen]:
                return Config.TILING[screen][desk_or_view]
            elif not isinstance(Config.TILING[screen], dict):
                return Config.TILING[screen]
        elif 'default' in Config.TILING:
            return Config.TILING['default']

        return Config.DEFAULTS['TILING']['default']

    @staticmethod
    def tilers(layout):
        if layout in Config.TILERS:
            return Config.TILERS[layout]

    @staticmethod
    def callbacks(num):
        if num in Config.CALLBACKS:
            return Config.CALLBACKS[num]
        if num in Config.DEFAULTS['CALLBACKS']:
            return Config.DEFAULTS['CALLBACKS'][num]
        return None

    # Special flag to enable/disable debugging
    #
    # PRIVACY NOTE: This may log the titles of
    # your windows!
    DEBUG = False



#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Maximal.py

A very simply layout algorithm that allows for full screen windows. For the most
part, the only useful commands are screen focus/put and focusing on the
previous/next window.
"""

from PyTyle.Tilers.TileDefault import TileDefault

class Maximal (TileDefault):
    #------------------------------------------------------------------------------
    # OVERLOADED INSTANCE METHODS
    #------------------------------------------------------------------------------
    
    #
    # We don't need to cycle here- previous/next commands do it for us already.
    #
    def _cycle(self):
        pass
    
    #
    # This Maximal layout simply resizes every window to the screen's width and
    # height. (We aren't actually "maximizing" it here. This seemed to flow a
    # bit better with tiling in general.
    #
    def _tile(self):
        x, y, width, height = self.screen.get_workarea()
        
        # set some vars...
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        
        masterWidth = width
        masterHeight = height
        masterY = y
        masterX = x
        
        slaveWidth = width
        slaveHeight = height
        slaveY = y
        slaveX = x
        
        # resize the master windows
        for master in masters:
            self.help_resize(master, masterX, masterY, masterWidth, masterHeight)
        
        # now resize the rest... keep track of heights/positioning
        for slave in slaves:
            self.help_resize(slave, slaveX, slaveY, slaveWidth, slaveHeight)   
    
    #
    # We want to disable the following methods. They are of no use for this
    # layout.
    # 
    def _master_increase(self, pixels = 50):
        pass
    
    def _master_decrease(self, pixels = 50):
        pass
    
    def _add_master(self):
        pass
    
    def _remove_master(self):
        pass
    
    def _make_active_master(self):
        pass
    
    def _win_master(self):
        pass
    
    def _switch_previous(self):
        pass
    
    def _switch_next(self):
        pass
    
    def _max_all(self):
        pass
    
    def _restore_all(self):
        pass
            
# You must have this line's equivalent for your tiling algorithm!
# This makes it possible to dynamically load tiling algorithms.
# (So that you may simply drop them into the Tilers directory,
# and add their name to the configuration- vini, vidi, vicci!)
CLASS = Maximal


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Probe.py

This file, along with Event.py, are the only two classes *aware*
of X. This means the rest of PyTyle could be yanked out and used
in other projects (YaTWM, anyone?). It also means you could rip
out this PROBE class and use it elsewhere.

Probe is a library class using Xlib to query the currently running
window manager (and X) for information about desktops, screens,
and windows.

This library includes everything from finding window sizes,
desktops, maximizing/restoring windows, to physical screen information
and a lot more.
"""

from Xlib.display import Display
from Xlib import X, XK, Xatom, Xutil, protocol
from Xlib.ext import xinerama
import sys, math

class Probe:
    #------------------------------------------------------------------------------
    # CONSTRUCTOR AND INSTANCE METHODS
    #------------------------------------------------------------------------------

    #
    # There should only be one Probe instance at any given time. Upon init,
    # instantiate the display object and fetch the root window. We also need to
    # listen to certain events on the root window:
    #    1. KeyPressMask - For mapping our hot keys
    #    2. SubstructureNotifyMask - For Create/Destroy window notification
    #    3. PropertyChangeMask - For desktop change notification
    #
    def __init__(self):
        self._display = Display()
        self._root = self.get_display().screen().root
        self._wm = ''
        self.determine_window_manager()
        self.get_root().change_attributes(event_mask = X.KeyPressMask | X.SubstructureNotifyMask | X.PropertyChangeMask)

    #
    # Alias to save some typing.
    # Display.intern_atom takes a string representation of an atom, and converts
    # it to its proper integer representation- which is what the X protocol uses.
    #
    def atom(self, name):
        return self.get_display().intern_atom(name)

    #
    # Finds the name of the current window manager.
    #
    def determine_window_manager(self):
        cid = self.get_root().get_full_property(self.atom("_NET_SUPPORTING_WM_CHECK"), 0)
        if not cid or not hasattr(cid, 'value'):
            return

        cid = cid.value[0]

        win = self.get_display().create_resource_object('window', cid)
        if not win:
            return

        name = win.get_full_property(self.atom("_NET_WM_NAME"), 0)
        if not name or not hasattr(name, 'value'):
            return

        self._wm = name.value.lower()

    #
    # Takes a string representation of a key and turns it into a key code for
    # use with grab_key. See Xlib/keysymdef/latin1.py and
    # Xlib/keysymdef/miscellany.py for available key code strings.
    #
    def generate_keycode(self, key):
        return self.get_display().keysym_to_keycode(XK.string_to_keysym(key))

    #
    # Takes a list of string modifiers and converts them to their proper mask
    # form. Essentially the equivalent of generate_keycode for key modifiers.
    # Currently the only modifiers supported are Shift, Ctrl, Alt, and Super.
    #
    def generate_modmask(self, mods):
        modmask = 0
        if len(mods) >= 1:
            for mod in mods:
                if mod == 'Shift':
                    modmask = modmask | X.ShiftMask
                elif mod == 'Ctrl':
                    modmask = modmask | X.ControlMask
                elif mod == 'Alt':
                    modmask = modmask | X.Mod1Mask
                elif mod == 'Super':
                    modmask = modmask | X.Mod4Mask
                else:
                    print >> sys.stderr, "Could not use modifier %s" % mod
        else:
            modmask = X.AnyModifier

        return modmask

    #
    # Queries the window manager for the currently active window. This is
    # what is *always* used to find the active window. We only rely on X events
    # to tell us when to use this (since the events can come from the root
    # window, and thus give us unknown id's for any given window).
    #
    # Note: It's possible that we won't have an active window.
    #
    def get_active_window_id(self):
        active = self.get_root().get_full_property(self.atom("_NET_ACTIVE_WINDOW"), 0)

        if hasattr(active, 'value'):
            return hex(active.value[0])
        else:
            return None

    #
    # Queries the window manager for the currently active desktop.
    #
    def get_desktop(self):
        return self.get_root().get_full_property(self.atom("_NET_CURRENT_DESKTOP"), 0).value[0]

    #
    # Queries the window manager for all available desktops. It also stores each
    # desktop name, which serves no function in PyTyle currently. (Good for easy
    # debugging though, I guess.) This is also where we query for our workarea
    # information such as the resolution, and the workarea accounting for panels
    # and/or docks. (See _NET_WM_STRUT and _NET_WM_STRUT_PARTIAL)
    #
    def get_desktops(self):
        info = {}
        #desktops = self.get_root().get_full_property(self.atom("_NET_DESKTOP_NAMES"), 0).value.split('\x00')[:-1]
        desktops = self.get_root().get_full_property(self.atom("_NET_NUMBER_OF_DESKTOPS"), 0).value[0]
        workarea = self.get_root().get_full_property(self.atom("_NET_WORKAREA"), 0).value
        resolution = self.get_root().get_full_property(self.atom("_NET_DESKTOP_GEOMETRY"), 0).value

        for i in range(desktops):
            info[i] = {
                       'id': int(i), 'x': workarea[0], 'y': workarea[1],
                       'width': workarea[2], 'height': workarea[3],
                       'resx': resolution[0], 'resy': resolution[1], 'name': 'None'
                       }

#        for i in range(len(desktops)):
#            desktop = desktops[i]
#            info[i] = {
#                       'id': int(i), 'x': workarea[0], 'y': workarea[1],
#                       'width': workarea[2], 'height': workarea[3],
#                       'resx': resolution[0], 'resy': resolution[1], 'name': desktop
#                       }

        return info

    #
    # Returns the current display object... Our connection to X.
    #
    def get_display(self):
        return self._display

    #
    # Returns the current root window.
    #
    def get_root(self):
        return self._root

    #
    # Tries to use xinerama to find multiple heads. The only thing xinerama
    # gives us are the x/y/width/height values of each screen. That means
    # we need to calculate which screen every window is on ourselves. You
    # can find this logic in the Screen class. Additionally, if xinerama
    # is not available, we simply build our own "Screen 0" with the
    # desktop geometry reported from the window manager. Don't forget, if
    # there is only one screen, the tiling algorithms will use the proper
    # workarea so that it accounts for panels/docks and the like, however,
    # if there is more than one screen, we need to manually specify docks
    # and panels in the configuration file. Why? Because it seems OpenBox
    # doesn't report the proper workarea back with multiple screens.
    #
    # Note: In fact, I have a feeling that this could get messy if there are
    # two panels (top/bottom, for instance)- so I'm going to enable the
    # config for docks/panels to take effect regardless of the number of
    # screens.
    #
    # Note 2: While PyTyle is written to handle an arbitrary number of screens,
    # I *unfortunately* don't have the capability of running triple
    # monitors- I hope that changes soon! If you do though, I would love to
    # hear how it's working (if at all). Email me: andrew@pytyle.com
    #
    def get_screens(self):
        ret = []

        if self.has_xinerama():
            screens = self.get_display().xinerama_query_screens().screens
            for i in range(len(screens)):
                screen = screens[i]
                ret.append({'id': i, 'x': screen.x, 'y': screen.y, 'width': screen.width, 'height': screen.height})

        if not ret:
            resolution = self.get_root().get_full_property(self.atom("_NET_DESKTOP_GEOMETRY"), 0).value
            ret = [{'id': 0, 'x': 0, 'y': 0, 'width': resolution[0], 'height': resolution[1]}]

        return ret

    #
    # Retrieves the current viewport. This is necessary for resizing
    # windows in managers like Compiz! Compiz thinks about windows
    # *relative* to the current viewport, so whenever we resize in a
    # window manager like that, we need to know the current viewport.
    #
    def get_viewport(self):
        viewport = self.get_root().get_full_property(self.atom("_NET_DESKTOP_VIEWPORT"), Xatom.CARDINAL)
        if viewport and hasattr(viewport, 'value'):
            return {'x': viewport.value[0], 'y': viewport.value[1]}
        return None

    #
    # Retrieves all available viewports. It uses some math trickery, but here's
    # the general gist:
    #    1. Ask the WM for the desktop geometry
    #    2. Ask the WM for the workarea.
    #    3. Divide the workarea into the desktop geometry, and floor it
    #        A. floor(desktop_width / workarea_width)
    #        B. floor(desktop_height / workarea_height)
    #        NOTE: We floor because the workarea might be smaller than the
    #        screen resolutions (struts).
    #    4. Calculations will give us number of horizontal and vertical
    #       viewports, respectively.
    #    5. Assigns id's: go vertical first, then wind back up to the
    #       next column.
    #
    def get_viewports(self):
        geom = self.get_root().get_full_property(self.atom("_NET_DESKTOP_GEOMETRY"), Xatom.CARDINAL)
        if self.is_compiz():
            workarea = self.get_root().get_full_property(self.atom("_NET_WORKAREA"), Xatom.CARDINAL)

            if not geom or not workarea:
                return None

            verts = int(math.floor(geom.value[1] / workarea.value[3]))
            horz = int(math.floor(geom.value[0] / workarea.value[2]))
            inc = 0
            viewports = []

            for v in range(verts):
                for h in range(horz):
                    viewports.append({
                                      'id': inc,
                                      'x': (geom.value[0] / horz) * h,
                                      'y': (geom.value[1] / verts) * v
                                      })
                    inc += 1
        else:
            viewports = [{
                          'id': 0,
                          'x': 0,
                          'y': 0
                          }]

        return viewports

    #
    # This will query the window manager for all necessary information for the
    # given window. This method takes a resource object created via
    # "create_resource_object"- or alternatively, straight from an event.
    # (Although, I don't recommend pulling a window straight from an event, as
    # it could be out of sync with PyTyle.) This method is called when a window
    # changes- we want to know its new state. (i.e., hidden, resized, desktop
    # or screen changed, etc.)
    #
    # Note: There is a lot that goes into this method, so please see the comments
    # below as well.
    #
    # Note 2: We don't calculate which screen the window is on from here. We do
    # that later (Window.load_window, essentially). However, this method supplies
    # what we need for that calculation- the window's x,y coordinates.
    #
    def get_window(self, win):
        # We don't really need the window name, but it's useful for debugging.
        # If a window doesn't have a name, or the window manager doesn't
        # listen to us, then we can still move on.
        winname = win.get_full_property(self.atom("_NET_WM_NAME"), 0)

        # Another way to find the window name.
        if not winname:
            winname = win.get_full_property(Xatom.WM_NAME, 0)
            if not winname: winname = ''

        if winname: winname = winname.value

        # Fetch the desktop that the window is on. We need this.
        #
        # *However* - This could change in the future if we want to support
        # viewports instead of desktops (*ahem* compiz *ahem*). We might need
        # to calculate the "desktop" based on the window's x,y coordinates
        # in the future. (Like we do for screens.)
        windesk = win.get_full_property(self.atom("_NET_WM_DESKTOP"), 0).value[0]

        # Fetch the window geometry- see the method below for more info.
        wingeom = self.get_window_geometry(win)

        # Extents are *hopefully* the window decoration sizes. PyTyle will
        # take these into account when sizing the windows. So far, support
        # seems pretty good for this from WM's.
        extents = win.get_full_property(self.atom("_NET_FRAME_EXTENTS"), 0)

        if not extents:
            extents = [0, 0, 0, 0]
        else:
            extents = extents.value

        # We use the normal hints to find the window's gravity. If the
        # gravity is static, then we need to change it to NorthWest at
        # some point if we expect the window to behave like we want it to.
        #
        # Note: This is obviously a hack. It would be better if we could
        # *account* for static gravity. Especially since static gravity
        # is set by the application itself, and if we don't let it have what
        # it wants, we might get unexpected behavior. So far so good, though.
        # If anyone has a better understanding of gravity than I do (that is,
        # beyond the usual man page), then I'd love to talk to you.
        norm_hints = win.get_wm_normal_hints()

        if norm_hints and norm_hints['win_gravity'] == X.StaticGravity:
            static = True
        else:
            static = False

        # So the transient will tell us a window's parent window. Why do we
        # care? Because if a top level window creates a child window (i.e.
        # a popup), then it will hopefully set that child window's transient
        # property as itself. If we come across transient windows, I don't
        # think we want to tile them. Sometimes the transient is set to the
        # root window, in which case it is obviously a window we want to tile.
        transient = win.get_wm_transient_for()

        if not transient or transient == self.get_root():
            popup = False
        else:
            popup = True

        # If a window is hidden, we don't want to tile it. (Or more importantly,
        # we don't want PyTyle to *think* there is a window that's there and
        # isn't.) Check for dock/panel/skip taskbar, etc...
        #
        # Note: We check both the "_NET_WM_STATE" (different from "WM_STATE" which
        # tells us about iconification and stuff) and the "_NET_WM_WINDOW_TYPE".
        state = win.get_full_property(self.atom("_NET_WM_STATE"), Xatom.ATOM)
        dock = win.get_full_property(self.atom("_NET_WM_WINDOW_TYPE"), Xatom.ATOM)
        hidden = False

        if state and (self.atom("_NET_WM_STATE_HIDDEN") in state.value or self.atom("_NET_WM_STATE_SKIP_TASKBAR") in state.value or self.atom("_NET_WM_STATE_SKIP_PAGER") in state.value):
            hidden = True
        if dock and (self.atom("_NET_WM_WINDOW_TYPE_DOCK") in dock.value or self.atom("_NET_WM_WINDOW_TYPE_TOOLBAR") in dock.value or self.atom("_NET_WM_WINDOW_TYPE_MENU") in dock.value or self.atom("_NET_WM_WINDOW_TYPE_SPLASH") in dock.value or self.atom("_NET_WM_WINDOW_TYPE_DIALOG") in dock.value):
            hidden = True

        # Construct the window data structure. This is passed to the
        # update_attributes method.
        return {
                 'id': hex(win.id),
                 'desktop': int(windesk),
                 'x': wingeom['x'], 'y': wingeom['y'],
                 'width': wingeom['width'], 'height': wingeom['height'],
                 'd_left': extents[0], 'd_right': extents[1],
                 'd_top': extents[2], 'd_bottom': extents[3],
                 'title': winname, 'class': win.get_wm_class(),
                 'static': static,
                 'popup': popup,
                 'hidden': hidden,
                 'xobj': win
                 }

    #
    # Simply fetchs a list of windows from the window manager. The list is
    # given to us as window id's- we then use that id to create a resource
    # object from which we can query.
    #
    def get_windows(self):
        info = {}
        windows = self.get_window_list()

        for window in windows:
            info[hex(window)] = self.get_window_by_id(window)

        return info

    #
    # Creates a window resource object from a given window id (decimal).
    #
    def get_window_by_id(self, window_id):
        win = self.get_display().create_resource_object("window", window_id)
        return self.get_window(win)

    #
    # It took me a little bit to figure this one out. So apparently, the
    # get_geometry window method returns coordinates that we don't care about.
    # (That is, they are relative to the root window?) So in order to
    # rectify this, we need to "translate" the x,y coordinates relative to
    # that root window. Not very intuitive at all. Kudos to wmctrl for showing
    # me the light here.
    #
    # Note: More testing has revealed different behavior for different window
    # managers here. For instance, with Compiz, we actually want the raw x,y
    # coordinates from get_geometry- translating them is bad! We end up with
    # some pretty big figures. (Viewports..?) I haven't investigated Compiz
    # thoroughly, but was able to get some minimal functionality by simply
    # removing the call to translate_coords.
    #
    def get_window_geometry(self, win):
        wingeom = win.get_geometry()

        wintrans = win.translate_coords(self.get_root(), wingeom.x, wingeom.y)
        wintrans.x = -wintrans.x
        wintrans.y = -wintrans.y

        # This is for compiz (and any other viewport-style WM?)...
        # looks like we don't need to translate
        if self.is_compiz():
            viewport = self.get_viewport()
            if viewport:
                wintrans = wingeom
                wintrans.x += viewport['x']
                wintrans.y += viewport['y']

        return {'x': wintrans.x, 'y': wintrans.y, 'width': wingeom.width, 'height': wingeom.height}

    #
    # Queries the window manager for a list of window id's. These window id's
    # are then used to create a window resource object from which we can query
    # for information about that specific window.
    #
    def get_window_list(self):
        return self.get_root().get_full_property(self.atom("_NET_CLIENT_LIST"), Xatom.WINDOW).value

    #
    # Returns the current window manager name.
    #
    def get_wm_name(self):
        return self._wm

    #
    # Another one that took forever to figure out. Grabbing a key *itself* is
    # pretty straight-forward. Unfortunately, I had number lock on. Ug. So
    # for each key we want to grab, we need to grab it normally, and then we
    # need to grab it three more times with the following masks:
    #    Mod2Mask (Number lock)
    #    LockMask (Caps lock)
    #    Mod2Mask | LockMask (Number lock and Caps lock)
    #
    # What about scroll lock? o_O
    #
    def grab_key(self, keycode, mask):
        self.get_root().grab_key(keycode, mask, 1, X.GrabModeAsync, X.GrabModeAsync)
        self.get_root().grab_key(keycode, mask | X.Mod2Mask, 1, X.GrabModeAsync, X.GrabModeAsync)
        self.get_root().grab_key(keycode, mask | X.LockMask, 1, X.GrabModeAsync, X.GrabModeAsync)
        self.get_root().grab_key(keycode, mask | X.Mod2Mask | X.LockMask, 1, X.GrabModeAsync, X.GrabModeAsync)

    #
    # Simply checks if the xinerama extension is enabled.
    #
    def has_xinerama(self):
        if self.get_display().has_extension('XINERAMA'):
            return True
        return False

    #
    # Checks to see if Compiz is running. It needs unique attention.
    #
    def is_compiz(self):
        if self.get_wm_name() == 'compiz':
            return True
        return False

    #
    # Reports if the window manager is running or not
    #
    def is_wm_running(self):
        try:
            PROBE.get_desktops()
        except:
            return False
        return True

    #
    # I was using this method originally in the main event loop, but found it
    # to be unnecessary after I polished up the get_window method and queried
    # for information about a window being a transient, along with whether or
    # not it was hidden. However, this method could still be potentially useful,
    # although it is not currently being used.
    #
    def is_popup(self, window):
        skip = window.get_full_property(self.atom("_NET_WM_STATE"), Xatom.ATOM)

        if skip and (self.atom("_NET_WM_STATE_MODAL") in skip.value or self.atom("_NET_WM_STATE_SKIP_TASKBAR") in skip.value or window.get_wm_transient_for()):
            return True
        return False

    #
    # Ungrabs a key (and all its modifiers). This allows us to dynamically reload
    # keybindings as PyTyle is running.
    #
    def ungrab_key(self, keycode, mask):
        self.get_root().ungrab_key(keycode, mask)
        self.get_root().ungrab_key(keycode, mask | X.Mod2Mask)
        self.get_root().ungrab_key(keycode, mask | X.LockMask)
        self.get_root().ungrab_key(keycode, mask | X.Mod2Mask | X.LockMask)

    #
    # Activates the given window. This will also pull it above all other
    # windows. Remember to flush.
    #
    def window_activate(self, win):
        win.set_input_focus(X.RevertToNone, X.CurrentTime)
        self.window_stackabove(win)
        self.get_display().flush()

    #
    # Attemps to remove window decorations, although I don't currently
    # recommend enabling the option to remove decorations. Why? Well,
    # because it will be quite difficult to tell which window is active,
    # since PyTyle isn't going to be drawing borders or anything.
    # (Well I suppose it could? I haven't investigated this yet, but
    # i'm not optimistic.)
    #
    # Props to devilspie for this one.
    #
    def window_add_decorations(self, win):
        # Doesn't seem to be working...
        #win.change_property(self.atom("_MOTIF_WM_HINTS"), self.atom("_MOTIF_WM_HINTS"), 32, [0x2, 0, 1, 0, 0])
        self._send_event(win, self.atom("_NET_WM_STATE"), [0, self.atom("_OB_WM_STATE_UNDECORATED")])
        self.get_display().flush()

    #
    # Simply closes the given window. This *functionality* isn't really
    # necessary for PyTyle, as it will detect when a window closes on
    # its own. I've included it for completeness (or if you don't have
    # a key binding to close a window).
    #
    def window_close(self, win):
        #win.destroy()
        self._send_event(win, self.atom("_NET_CLOSE_WINDOW"), [X.CurrentTime])
        self.get_display().flush()

    #
    # This sets up the event mask on the given window. This will tell the
    # X server to send us only the events we're interested in (however,
    # we still get a boat load more than what we really care about).
    # They are as follows:
    #    FocusChangeMask    Whenever a window changes focus. We only use
    #                       this to detect FocusIn events, and I'm thinking
    #                       about using PropertyChange on the root window
    #                       instead of this.
    #    StructureNotifyMask    For when a window changes x/y/width/height.
    #    PropertyChangeMask     For when a window changes desktops.
    #
    # Note: See Event.py for more information on events.
    #
    def window_listen(self, win):
        win.change_attributes(event_mask = (X.FocusChangeMask | X.StructureNotifyMask | X.PropertyChangeMask))

    #
    # Simply maximizes a window. We must send a client message event to the
    # root window for this. (Or any other _NET_WM_STATE_* property.)
    #
    def window_maximize(self, win):
        self._send_event(win, self.atom("_NET_WM_STATE"), [1, self.atom("_NET_WM_STATE_MAXIMIZED_VERT"), self.atom("_NET_WM_STATE_MAXIMIZED_HORZ")])
        #win.change_property(self.atom("_NET_WM_STATE"), Xatom.ATOM, 32, [1, self.atom("_NET_WM_STATE_MAXIMIZED_VERT"), self.atom("_NET_WM_STATE_MAXIMIZED_HORZ")])
        self.get_display().flush()

    #
    # See window_add_decorations.
    #
    def window_remove_decorations(self, win):
        # Doesn't seem to be working...
        #win.change_property(self.atom("_MOTIF_WM_HINTS"), self.atom("_MOTIF_WM_HINTS"), 32, [0x2, 0, 0, 0, 0])
        self._send_event(win, self.atom("_NET_WM_STATE"), [1, self.atom("_OB_WM_STATE_UNDECORATED")])
        self.get_display().flush()

    #
    # Attempts to set window gravity to NorthWest. So far this has been
    # working well, although changing a window's gravity is a hack. See
    # get_window for more information on this gravity stuff as it
    # relates to PyTyle.
    #
    def window_remove_static(self, window):
        window.set_wm_normal_hints(
                                   flags = Xutil.PWinGravity,
                                   win_gravity = X.NorthWestGravity
                                   )
        self.get_display().flush()

    #
    # This simply "unmaximizes" or "restores" a window. We need to do this
    # every time we resize a window because it could have been maximized
    # by the user (which then could not be resized).
    #
    def window_reset(self, win):
        self._send_event(win, self.atom("_NET_WM_STATE"), [0, self.atom("_NET_WM_STATE_MAXIMIZED_VERT"), self.atom("_NET_WM_STATE_MAXIMIZED_HORZ")])
        #win.change_property(self.atom("_NET_WM_STATE"), Xatom.ATOM, 32, [0, self.atom("_NET_WM_STATE_MAXIMIZED_VERT"), self.atom("_NET_WM_STATE_MAXIMIZED_HORZ")])
        self.get_display().flush()

    #
    # Resizes the window with the given x/y/width/height pixel values.
    # Don't forget to flush after and reset the window before.
    #
    def window_resize(self, win, x, y, width, height):
        self.window_reset(win)

        # This is for compiz (and any other viewport-style WM?)...
        # looks like we don't need to translate
        if self.is_compiz():
            viewport = self.get_viewport()
            if viewport:
                x -= viewport['x']
                y -= viewport['y']

        win.configure(x=x, y=y, width=width, height=height)
        self.get_display().flush()

    #
    # Puts window at the top of the stack.
    #
    def window_stackabove(self, win):
        win.configure(stack_mode=X.Above)

    #
    # Puts window at the bottom of the stack.
    #
    def window_stackbelow(self, win):
        win.configure(stack_mode=X.Below)

    #
    # Stop listening to a window. PyTyle does not currently use this,
    # but I was toying with it when I was hitting positive feedback
    # in my main event loop. Yuck.
    #
    def window_unlisten(self, win):
        win.change_attributes(event_mask = 0)


    #------------------------------------------------------------------------------
    # PRIVATE INSTANCE HELPER METHODS
    #------------------------------------------------------------------------------

    #
    # Another tricky one to figure out- this will allow you to send
    # a client message to the root window (necessary for removing
    # decorations, maximizing, etc).
    #
    # Props to PyPanel for this little snippet.
    #
    def _send_event(self, win, ctype, data, mask=None):
        data = (data + ([0] * (5 - len(data))))[:5]
        ev = protocol.event.ClientMessage(window=win, client_type=ctype, data=(32, (data)))
        self.get_root().send_event(ev, event_mask=X.SubstructureRedirectMask)

#
# Instantiate the PROBE instance. This is what we import.
#
PROBE = Probe()


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Viewport.py

The Viewport class will handle all viewport related functionality. For instance,
it provides methods to detect if a given window is on its viewport.
"""

from PyTyle.State import State
from PyTyle.Probe import PROBE

from PyTyle.Screen import Screen

class Viewport:
    #------------------------------------------------------------------------------
    # CONSTRCUTOR AND VIEWPORT RELATED ATTRIBUTES/METHODS
    #------------------------------------------------------------------------------ 
    
    #
    # Initialization requies a desktop and attributes fetched from X. We also make
    # sure that it starts out with:
    #    1. Tiling disabled
    #    2. Untiled state
    #    3. No tiler
    #
    def __init__(self, desktop, attrs):
        self.update_attributes(attrs)
        self._SCREEN = None
        self.desktop = desktop 
        
        if PROBE.is_compiz():
            self.width = self.desktop.width
            self.height = self.desktop.height
        else:
            self.width = self.desktop.resx
            self.height = self.desktop.resy
        
        self.screens = {}
        self.load_screens()
    
    #
    # Takes a pair of x,y coordinates and tells us whether they are in the
    # viewport's grid. Also, take special care for windows with a negative x,y
    # position- simply say that it is on the viewport with x,y coordinates
    # equal to 0.
    #
    def is_on_viewport(self, x, y):
        if x >= self.x and y >= self.y and x < (self.x + self.width) and y < (self.y + self.height):
            return True
        
        # off the screen?
        if (x < 0 or y < 0) and self.x == 0 and self.y == 0:
            return True
        
        return False
    
    #
    # Probes X (using xinerama) for all available screens. For every viewport,
    # an instance of each screen is newly created. (So the total number of
    # "screens" in PyTyle is # of physical screens * viewports * desktops.) We 
    # do *not* queue screens for tiling here.
    #
    def load_screens(self):
        screens = PROBE.get_screens()
        for screen in screens:
            obj = Screen(self, screen)
            obj.x += self.x
            obj.y += self.y
            self.screens[screen['id']] = obj
                
    #
    # Updates viewport with attributes fetched from X.
    #
    def update_attributes(self, attrs):
        self.id = attrs['id']
        self.x = attrs['x']
        self.y = attrs['y']
        
    #
    # String representation of the current screen. See also the string
    # representations of desktop and window.
    #                
    def __str__(self):
        return 'Viewport ' + str(self.id) + ' - [ID: ' + str(self.id) + ', X: ' + str(self.x) + ', Y: ' + str(self.y) + ', WIDTH: ' + str(self.width) + ', HEIGHT: ' + str(self.height) + ', DESKTOP: ' + str(self.desktop.id) + ']'


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Event.py

This class, along with Probe, are the only two classes interfacing
with X in PyTyle.

Event handles all event related information. Each instance is a
separate event sent from X.

Most of the function of this class is to encapsulate event information.
"""

from PyTyle.Probe import PROBE
from PyTyle.Debug import DEBUG
from Xlib import X

class Event:    
    #------------------------------------------------------------------------------
    # CONSTRUCTOR AND INSTANCE METHODS
    #------------------------------------------------------------------------------
    
    #
    # Each instance represents one event. Upon initialization, we grab that event
    # from the X server.
    #
    def __init__(self):
        self._event = PROBE.get_display().next_event()
        
    #
    # Fetches the window id from the event. We have to convert it to a long and
    # then a hex so that it matches up with our dictionary of windows (the State).
    #
    def get_window_id(self):
        if self._event and hasattr(self._event, 'window'):
            return hex(long(self._event.window.id))
        return 0
    
    #
    # Fetches the key code from the event object. Make sure it's a KeyPress event.
    #
    def get_keycode(self):
        if self._event and self._event.type == X.KeyPress:
            return self._event.detail
        return None
    
    #
    # The key "state" is a bit mask of which modifiers were pressed when the key
    # was hit. See get_masks.
    #
    def get_key_state(self):
        if self._event and self._event.type == X.KeyPress:
            return self._event.state
        return 0
        
    #
    # This method detects exactly which modifiers are represented in the current
    # key state. Supported modifiers are as follows:
    #    ShiftMask - Shift key
    #    ControlMask - Control key (either)
    #    Mod1Mask - Alt key (either)
    #    Mod4Mask - Super
    #
    def get_masks(self):
        if not self._event or not self.is_keypress():
            return None
        
        ret = 0
        state = self.get_key_state()
        if X.ShiftMask & state:
            ret = ret | X.ShiftMask
        if X.ControlMask & state:
            ret = ret | X.ControlMask
        if X.Mod1Mask & state:
            ret = ret | X.Mod1Mask
        if X.Mod4Mask & state:
            ret = ret | X.Mod4Mask
            
        if not ret:
            ret = X.AnyModifier
        
        return ret
    
    #
    # Reports whether this event is an active window changing event.
    #
    def is_active_change(self):
        if self._event and self._event.type == X.PropertyNotify and self._event.atom == PROBE.atom("_NET_ACTIVE_WINDOW"):
            return True
        return False
    
    #
    # Reports whether this event is a desktop changing event. This event is
    # sent from the root window. We use this to update the active window.
    # Although I am thinking of changing to listening for the
    # _NET_ACTIVE_WINDOW instead. Will investigate more later.
    #
    def is_desktop_change(self):
        if self._event and self._event.type == X.PropertyNotify and (self._event.atom == PROBE.atom("_NET_CURRENT_DESKTOP") or self._event.atom == PROBE.atom("_NET_DESKTOP_VIEWPORT")):
            return True
        return False
    
    #
    # Reports whether this is a clientMessage.
    #
    def is_client_message(self):
        if self._event and self._event.type == X.ClientMessage and self._event.client_type == PROBE.atom("_PYTYLE_REMOTE"):
            return True
        return False

    def get_client_payload(self):
        if not self._event or not self.is_client_message():
            return None
        (format,data) = self._event.data
        if format==32:
            return data[0]
        else:
            return None
    #
    # Reports whether the current event is a focus *in* event. (We don't
    # care about focus *out* right now.) We also make sure that this is
    # a normal focus event (otherwise we get flooded with crap we don't
    # need.)
    #
    def is_focus_in(self):
        if self._event and self._event.type == X.FocusIn and self._event.mode == X.NotifyNormal:
            return True
        return False
        
    #
    # Reports whether the current event is a key press or not.
    #
    def is_keypress(self):
        if self._event and self._event.type == X.KeyPress:
            return True
        return False
    
    #
    # Reports whether the screen setup has changed by checking the
    # _NET_DESKTOP_GEOMETRY property.
    #
    def is_screen_change(self):
        if self._event and self._event.type == X.PropertyNotify and (self._event.atom == PROBE.atom("_NET_DESKTOP_GEOMETRY") or self._event.atom == PROBE.atom("_NET_NUMBER_OF_DESKTOPS")):
            return True
        return False
    
    #
    # Reports whether the window's state has changed. If a window's state
    # changes (i.e., it was hidden or something), then we need to refresh
    # its information and tell the screen it needs to be retiled.
    #
    def is_state_change(self):
        if self._event and self._event.type == X.PropertyNotify and self._event.atom == PROBE.atom("WM_STATE"):
            return True
        return False
    
    #
    # Reports whether the window manager's client list has changed or not.
    # Useful for detecting add/removal of windows.
    #
    def is_windowlist_change(self):
        if self._event and self._event.type == X.PropertyNotify and self._event.atom == PROBE.atom("_NET_CLIENT_LIST"):
            return True
        return False
    
    #
    # Reports whether the event is a window change or not. We want to know
    # if the window changes whenever it is resized/moved (ConfigureNotify),
    # or when its _NET_WM_DESKTOP property changes (PropertyNotify). This
    # allows for us to drag windows to and from tiling screens (by using
    # the mouse, keyboard, desktop switch key, etc).
    #
    def is_window_change(self):
        if self._event and ((self._event.type == X.ConfigureNotify and self._event.event != PROBE.get_root()) or (self._event.type == X.PropertyNotify and self._event.atom == PROBE.atom("_NET_WM_DESKTOP"))):
            return True
        return False
    
    #
    # Reports whether we are creating a window or not. This will initiate
    # a scan for new windows in the client list. Why do we scan? Because
    # don't actually use the window object that comes attached to with
    # this event, because its ID might not line up with the one we
    # normally receive in PyTyle (i.e., from the _NET_CLIENT_LIST).
    #
    def is_window_create(self):
        if self._event and self._event.type == X.CreateNotify:
            return True
        return False
    
    #
    # Reports whether a window was destroyed or not. This will initiate
    # a scan of all current windows and the State to see if there are
    # any windows in PyTyle that aren't in the client list. We will
    # then remove it. Like when we create a window, we don't use the
    # window object attached to this event because it is unreliable
    # for our purposes.
    #
    def is_window_destroy(self):
        if self._event and self._event.type == X.DestroyNotify:
            return True
        return False
    
    #
    # Reports whether the workarea has changed (i.e., a dock/panel has
    # been added to the screen).
    #
    def is_workarea_change(self):
        if self._event and self._event.type == X.PropertyNotify and self._event.atom == PROBE.atom("_NET_WORKAREA"):
            return True
        return False


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Debug.py

A very simple class to log messages.
"""

import time, os, sys
from PyTyle.Config import Config
from PyTyle.State import State

class Debug:
    #------------------------------------------------------------------------------
    # CONSTRCUTOR AND INSTANCE METHODS
    #------------------------------------------------------------------------------
    
    
    #
    # Simply opens the log file. Keep the log file going.
    #
    def __init__(self, filename):
        self._log = sys.stderr
        if Config.DEBUG:
            self._log = open(filename, 'a+')
            print >> self._log, '\n\n', '---------------------------------'
            self.write('PyTyle started')
            
    #
    # Writes a message to the log file
    #
    def write(self, msg):
        if not self._log:
            return
        
        t = time.localtime()
        write = '%d/%d/%d at %d:%d:%d:    %s' % (t.tm_mon, t.tm_mday, t.tm_year, t.tm_hour, t.tm_min, t.tm_sec, msg)
        print >> self._log, write 
        self._log.flush()
        
DEBUG = Debug(os.getenv('HOME') + '/pytyle.log')


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
State.py

Keeps a record of everything going on in the outside world. It keeps track
of all windows that PyTyle knows about, along with all desktops (but not
all screens- no need). It also keeps a record of the current desktop. Since
each desktop keeps a record of its current screen, and each screen keeps a
record of its current window, State provides us with easy access to the active
window while also giving us its desktop and screen.

Additionally, State keeps track of which screens need tiling. This serves as
a queue, and periodically, after certain things have happened (eg., a window
was hidden), a screen will be queued up.

State also serves to initialize hot keys and scans for new windows.
"""

from PyTyle.Config import Config
from PyTyle.Probe import PROBE

class State:
    #------------------------------------------------------------------------------
    # CLASS VARIABLES
    #------------------------------------------------------------------------------ 
    
    #
    # Keeps track of the currently active desktop.
    #
    _DESKTOP = None
    
    #
    # Keeps a record of all instantiated desktops.
    #
    _DESKTOPS = {}
    
    #
    # Keeps a mapping of keys to tiling actions
    #
    _DISPATCHER = {}
    
    #
    # Tells us whether we need to reload the config file.
    #
    _RELOAD = False
    
    #
    # Queue of screens to tile. It's flushed at the start of each event loop
    # iteration.
    #
    _TO_TILE = []
    
    #
    # Keeps a record of all instantiated windows.
    #
    # Note: If a window is loaded in this dict, it does *not* mean it will
    # definitely be tiled. Which windows are actually tiled is up to the TileStorage
    # class to determine. (That is where the window filter is used.)
    #
    _WINDOWS = {}
        
    
    #------------------------------------------------------------------------------
    # STATIC METHODS
    #------------------------------------------------------------------------------
    
    #
    # Adds a desktop to the state.
    #
    @staticmethod
    def add_desktop(desktop):
        State._DESKTOPS[desktop.id] = desktop
        
    #
    # Adds a window to the state.
    #
    @staticmethod
    def add_window(window):
        State._WINDOWS[window.id] = window
        
    #
    # Removes a window from the state.
    #
    @staticmethod
    def delete_window(window):
        del State._WINDOWS[window.id]
        
    #
    # Removes a screen from queue. Only used when flushing the queue
    # at the start of each event iteration.
    #
    @staticmethod
    def dequeue_screen():
        return State._TO_TILE.pop(0)
    
    #
    # Unsets the flag to reload the config file.
    #
    @staticmethod
    def did_reload():
        State._RELOAD = False
    
    #
    # Sets a flag to have PyTyle reload the config file.
    #
    @staticmethod
    def do_reload():
        State._RELOAD = True
    
    #
    # Retrieves the current desktop.
    #
    @staticmethod
    def get_desktop():
        return State._DESKTOP
    
    #
    # Retrieves the desktops in the state.
    #
    @staticmethod
    def get_desktops():
        return State._DESKTOPS
    
    #
    # Retrieves the dispatcher.
    #
    @staticmethod
    def get_dispatcher():
        return State._DISPATCHER
    
    #
    # Retrieves the windows in the state.
    #
    @staticmethod
    def get_windows():
        return State._WINDOWS
    
    #
    # Retrieves the name of the currently running window manager.
    #
    @staticmethod
    def get_wm_name():
        return PROBE.get_wm_name()
    
    #
    # Tells us whether we need to reload the config file.
    #
    @staticmethod
    def needs_reload():
        return State._RELOAD
        
    #
    # Reports whether the tiling queue is empty or not.
    #
    @staticmethod
    def queue_has_screens():
        return True if State._TO_TILE else False
    
    #
    # Adds a screen to the tiling queue. (You shouldn't use this method
    # directly to queue up a screen, but rather, the "needs_tiling" method
    # in the Screen class.)
    #
    @staticmethod
    def queue_screen(screen):
        State._TO_TILE.append(screen)
        
    #
    # Simply ties a key code to a callback method in the Tile class. Valid key codes
    # can be found in the documentation provided. (But are based on the key symbols
    # defined in Xlib. Probably in Xlib/keysymdef/latin1.py and Xlib/keysymdef/miscellany.py)
    # 
    @staticmethod
    def register_hotkey(keycode, mask, callback):
        if not keycode in State._DISPATCHER:
            State._DISPATCHER[keycode] = {}
        State._DISPATCHER[keycode][mask] = callback
        
    #
    # Registers all the key bindings specified in the configuration file. Currently,
    # only Shift, Ctrl, Alt, and/or Super are supported are modifier keys. See also,
    # State.register_hotkey and the documentation for a more complete description
    # of key combinations.
    #
    @staticmethod
    def register_hotkeys():
        for mapping in Config.KEYMAP:
            callback = Config.KEYMAP[mapping]
            
            codes = mapping.split('-')
            mods = codes[:-1]
            key = codes[-1]
            
            # No key?
            if not key:
                print >> sys.stderr, "Could not map %s to %s" % (mapping, callback) 
                continue
            
            # generate key code and mod mask...
            keycode = PROBE.generate_keycode(key)
            modmask = PROBE.generate_modmask(mods)
            
            # Tell X we want to hear about it when this key is pressed...
            try:
                PROBE.grab_key(keycode, modmask)
            except:
                print "Nada:", callback
            
            # Finally register the key with the dispatcher...            
            State.register_hotkey(keycode, modmask, callback)
        
    #
    # Simply probes for the currently active window, and updates the currently
    # active desktop, screen, and window accordingly.
    #
    @staticmethod
    def reload_active(active = None, force = False):
        if not active: 
            activeid = PROBE.get_active_window_id()
        else:
            activeid = active.id
            
        # Check to make sure we need to probe for anything...
        if not force and activeid and State._DESKTOP and State._DESKTOP._VIEWPORT and State._DESKTOP._VIEWPORT._SCREEN:
            current = State._DESKTOP._VIEWPORT._SCREEN.get_active()
            if current and current.id == activeid:
                return
            
        State._DESKTOP = State.get_desktops()[PROBE.get_desktop()]        
                
        if not activeid:
            if not State._DESKTOP._VIEWPORT:
                State._DESKTOP._VIEWPORT = State._DESKTOP.viewports[0]
                
            if not State._DESKTOP._VIEWPORT._SCREEN:
                State._DESKTOP._VIEWPORT._SCREEN = State._DESKTOP._VIEWPORT.screens[0]
        else:
            for viewport in State._DESKTOP.viewports.values():
                for screen in viewport.screens.values():
                    if activeid in screen.windows:
                        State._DESKTOP._VIEWPORT = viewport
                        State._DESKTOP._VIEWPORT._SCREEN = screen
                        State._DESKTOP._VIEWPORT._SCREEN.set_active(screen.windows[activeid])
                        break
            
    #
    # Similar to scan_new_windows, except it returns all windows reported by
    # the window manager. We used this when we get a DestroyNotify event: iterate
    # over the State windows and see which ones aren't in this list and delete
    # them.
    #
    @staticmethod
    def scan_all_windows():
        ret = []
        windows = PROBE.get_window_list()
        for window in windows:
            ret.append(window) 
                
        return ret
    
    #
    # There are problems with using the window given via the CreateNotify event
    # type, and thus, PyTyle currently listens for that event and scans for new
    # windows manually.
    #
    @staticmethod
    def scan_new_windows():
        ret = []
        windows = PROBE.get_window_list()
        for window in windows:
            if hex(window) not in State.get_windows():
                ret.append(window) 
                
        return ret
    
    #
    # UN-registers all the key bindings specified in the configuration file. This
    # allows us to dynamically change key bindings as PyTyle is running.
    #
    @staticmethod
    def unregister_hotkeys():
        for mapping in Config.KEYMAP:
            callback = Config.KEYMAP[mapping]
            
            codes = mapping.split('-')
            mods = codes[:-1]
            key = codes[-1]
            
            # No key?
            if not key:
                print >> sys.stderr, "Could not map %s to %s" % (mapping, callback) 
                continue
            
            # generate key code and mod mask...
            keycode = PROBE.generate_keycode(key)
            modmask = PROBE.generate_modmask(mods)
            
            # Tell X we want to hear about it when this key is pressed...
            try:
                PROBE.ungrab_key(keycode, modmask)
            except:
                print "Nada:", callback
            
        # And finally reset the dispatcher...
        State._DISPATCHER = {}
    
    #
    # Wipes the current state. Useful for when the screen orientation changes.
    #
    @staticmethod
    def wipe():
        State._DESKTOP = None
        State._WINDOWS = {}
        State._DESKTOPS = {}
        State._TO_TILE = []


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Screen.py

The Screen class will handle all screen related functionality. For instance,
it provides methods to detect if a given window is on its screen. Also, it
will keep track of its state; like if it needs to be tiled, has tiling
enabled/disable, and of course, its current tiling algorithm.
"""

from PyTyle.Config import Config
from PyTyle.State import State
from PyTyle.Debug import DEBUG

from PyTyle.Window import *

class Screen:
    #------------------------------------------------------------------------------
    # CONSTRCUTOR AND SCREEN RELATED ATTRIBUTES/METHODS
    #------------------------------------------------------------------------------

    #
    # Initialization requies a desktop and attributes fetched from X. We also make
    # sure that it starts out with:
    #    1. Tiling disabled
    #    2. Untiled state
    #    3. No tiler
    #
    def __init__(self, viewport, attrs):
        self.update_attributes(attrs)
        self._active = None
        self._tile = None
        self._tiled = False
        self._tiling = Config.misc('global_tiling')
        self.windows = {}
        self.viewport = viewport

    #
    # Disables tiling for this screen. Essentially, this is called when the
    # untile method is called via a key press. Meaning, if you programmatically
    # call untile, it will not disable tiling. (It will however, resize the
    # windows back to their original state- is that good behavior?)
    def disable_tiling(self):
        self._tiling = False

    #
    # Tells the screen that PyTyle can now touch it. Currently, a screen can
    # *only* be enabled for tiling using a key binding tied to the tile method.
    # Even if the tile method is called programmatically (i.e., not instantiated
    # via a key press), it won't tile if tiling isn't enabled.
    #
    def enable_tiling(self):
        self._tiling = True

    #
    # Fetches the currently active window for *this* screen. That means that
    # the currently active window for this screen may not be *THE* active
    # window. Be careful about that.
    #
    # Logic: The screen is in charge of reporting back an active window,
    # and also saying when there isn't one. There are four possibilities:
    #    1. If there is no active window set, and there are no windows,
    #       then return nothing.
    #    2. If there are windows, and no active window, then arbitrarily
    #       select the first window in the screen window storage to make
    #       active. Also, if there are windows and the currently active window
    #       is no longer on this screen, then similarly choose an arbitrary
    #       window to make active.
    #    3. If there are no windows but there is an active window
    #       and it's no longer on the screen, then return nothing.
    #    4. Finally, if it's none of the above, then return the currently
    #       set active window.
    #
    def get_active(self):
        wins = self.get_tiler().storage.get_all()
        if not self._active and not wins:
            self._active = None
        elif wins and (not self._active or self._active.hidden or self._active.screen.id != self.id or not self._active.lives()):
            self._active = wins[0]
        elif not wins and self._active.screen.id != self.id:
            self._active = None

        return self._active

    #
    # Fetches the current tiling algorithm. The tiling algorithm *must* be
    # a subclass of Tile. (Which does a lot of the grunt work for you.)
    #
    def get_tiler(self):
        return self._tile

    #
    # Fetches the workarea of the current screen. Essentially, this has worked
    # quite well in my experience with just one monitor. However, when we add
    # monitors, the window manager doesn't seem to report desktop dimensions
    # back correctly, in a way that allows PyTyle to see where docks/panels are.
    # My solution to this was to provide a manual override in the configuration
    # file for multiple screens. It isn't pretty, but it's what I've got so far.
    #
    # Note: This is used in the core part of the tiling algorithms to place
    # the windows. If you have some fancy setup in your workarea, this is the
    # place to tinker! (If the configuration file isn't enough power.)
    #
    def get_workarea(self):
        # If we have one screen, look for a "Screen 0" config
        # and use it if it exists...
        if len(self.viewport.screens) == 1:
            if 0 in Config.WORKAREA:
                x = self.x + Config.workarea(0, 'left')
                y = self.y + Config.workarea(0, 'top')
                height = self.height - Config.workarea(0, 'bottom') + Config.workarea(0, 'top')
                width = self.width - Config.workarea(0, 'right') + Config.workarea(0, 'left')
            else:
                if PROBE.is_compiz():
                    x = self.viewport.x
                    y = self.viewport.y
                else:
                    x = self.viewport.desktop.x
                    y = self.viewport.desktop.y

                height = self.viewport.desktop.height
                width = self.viewport.desktop.width
                print x, y, width, height
        else:
            x = self.x
            y = self.y
            height = self.height
            width = self.width

            # Factor in manual docks...
            x += Config.workarea(self.id, 'left')
            y += Config.workarea(self.id, 'top')
            height -= Config.workarea(self.id, 'bottom') + Config.workarea(self.id, 'top')
            width -= Config.workarea(self.id, 'right') + Config.workarea(self.id, 'left')

        return (x, y, width, height)

    #
    # This is used whenever the screen's tiler calls the "tile" method. It
    # tells the screen that everything is find and dandy, like sour candy.
    #
    def got_tiling(self):
        self._tiled = True

    #
    # Reports whether the screen is currently tiled. This is essentially
    # used by the Tile dispatcher. If we are coming into the tile method,
    # and the screen reports that it isn't tiled, then that means something
    # on the screen triggered it and told it needed to be re-tiled. And
    # thus, we re-tile and set this attribute to true.
    #
    def is_tiled(self):
        return self._tiled

    #
    # Reports whether this screen is in tiling mode or not.
    #
    def is_tiling(self):
        return self._tiling

    #
    # This tells the screen that something happened and it therefore needs
    # to be re-tiled. So, this should be called any time something changes
    # on a screen (new window, window destroyed, etc). When it's called,
    # it will also queue the screen into the State so that PyTyle will
    # re-tile the screen when it gets a chance.
    #
    # Note: We need more than just a queue- and that's why this attribute
    # exists. The screen itself needs to be aware if it's in a proper
    # state or not so that the tiler knows if it should reload its storage.
    #
    def needs_tiling(self):
        self._tiled = False
        State.queue_screen(self)

    #
    # Simply sets the active window. No questions asked.
    #
    def set_active(self, window):
        self._active = window

    #
    # Sets the tiler of this screen. A tiler *must* be a subclass of Tile.
    # To switch tiling layouts (if the default layout cycler isn't enough),
    # it's as simple as calling this method, followed by a call to the
    # _reset method in the Tile class.
    #
    def set_tiler(self, tile):
        self._tile = tile(self)

    #
    # This will add a new window to the screen. It will also take care of
    # updating the State for you. Additionally, it will tell itself that it
    # needs to be re-tiled.
    #
    def add_window(self, window):
        State.add_window(window)
        self.windows[window.id] = window
        self.needs_tiling()

    #
    # Opposite of add_window. It will still tell itself it needs to be
    # re-tiled.
    #
    def delete_window(self, window):
        State.delete_window(window)
        del self.windows[window.id]
        self.needs_tiling()

    #
    # Tales a pair of x,y coordinates and the width and height of a window,
    # and returns True if the window is contained within the current screen.
    # False otherwise.
    #
    def is_in_screen(self, x, y, width, height):
        if not self.is_on_screen(x, y):
            return False

        if width >= self.width or height >= self.height:
            return False

        return True

    #
    # Takes a pair of x,y coordinates and tells us whether they are in the
    # screen's grid. Also, take special care for windows with a negative x,y
    # position- simply say that it is on the screen with x,y coordinates
    # equal to 0.
    #
    def is_on_screen(self, x, y):
        if x >= self.x and y >= self.y and x < (self.x + self.width) and y < (self.y + self.height):
            return True

        # off the screen?
        if (x < 0 or y < 0) and self.x == 0 and self.y == 0:
            return True

        return False

    #
    # Updates screen with attributes fetched from X. This is actually everything
    # that the xinerama extension gives us. I'm not sure if screen ids remain
    # the same for each monitor on boot, although I imagine they don't. I'll worry
    # about that later. It's been fine so far. (PyTyle will still function
    # perfectly if id's are rearranged- but the key bindings will be reversed. Not
    # very user-friendly.)
    #
    def update_attributes(self, attrs):
        self.id = attrs['id']
        self.x = attrs['x']
        self.y = attrs['y']
        self.width = attrs['width']
        self.height = attrs['height']

    #
    # String representation of the current screen. See also the string
    # representations of desktop, viewport, and window.
    #
    def __str__(self):
        return 'Screen ' + str(self.id) + ' - [ID: ' + str(self.id) + ', X: ' + str(self.x) + ', Y: ' + str(self.y) + ', WIDTH: ' + str(self.width) + ', HEIGHT: ' + str(self.height) + ', VIEWPORT: ' + str(self.viewport.id) + ', DESKTOP: ' + str(self.viewport.desktop.id) + ']'



#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
TileStorage.py

My attempt at abstracting the storage of windows for tiling algorithms.
Namely, the handling of window ordering so that we have a data structure
that closely resembles that which you see on the screen. Additionally, it
will handle the management of masters and slaves. Tiling algorithms need
only be aware of the masters and slaves, and shouldn't have to manage which
windows become masters or slaves (unless you need to).

This is also where we run our filter specified in the configuration.
(Should this be moved to load_window?) This is also where we decide to
show/hide hidden windows. Namely, you can "add" a window to the tiling
storage, and it's possible that it won't be added. That is intentional.

Remember, if a window isn't in the storage, it will *not* be tiled. While
it will have a place in PyTyle, the tiling algorithm will not be able
to see it.
"""

class TileStorage:
    #------------------------------------------------------------------------------
    # CONSTRUCTOR AND INSTANCE METHODS
    #------------------------------------------------------------------------------
    
    #
    # Will start the master count at 1, and initialize the master and slave
    # *ordered* lists.
    #
    def __init__(self):
        self._master_count = 1
        self._masters = []
        self._slaves = []
        
    #
    # Adds a window to the storage. This will detect if that window should
    # be a slave or a master based on the current master count and how
    # many masters are currently loaded in the storage.
    #
    def add(self, window):
        if window.hidden: return
        
        if len(self._masters) < self.get_master_count():
            if window.id in self.get_slaves_by_id():
                self._remove_slave(window)
            self._add_master(window)
        else:
            self._add_slave(window)
            
    def add_top(self, window):
        if window.hidden: return
        
        if len(self._masters) < self.get_master_count():
            if window.id in self.get_slaves_by_id():
                self._remove_slave(window)
            self._add_top_master(window)
        else:
            self._add_top_slave(window)
            
    def add_bottom(self, window):
        if window.hidden: return
        
        if len(self._masters) < self.get_master_count():
            if window.id in self.get_slaves_by_id():
                self._remove_slave(window)
            self._add_bottom_master(window)
        else:
            self._add_bottom_slave(window)
        
    #
    # Decrements the number of masters allowed for this tiler. It cannot
    # go below 0 for obvious reasons (but can be equal to 0).
    #
    def dec_master_count(self):
        if self._master_count == 0: return
        self._master_count -= 1
    
    #
    # Returns all windows currently in the storage.
    #    
    def get_all(self):
        return self.get_masters() + self.get_slaves()
    
    #
    # Returns all windows currently in the storage by their id.
    #
    def get_all_by_id(self):
        return self.get_masters_by_id() + self.get_slaves_by_id()
    
    #
    # Returns all masters currently in the storage. It will
    # conveniently extract the windows from the tupling storage
    # mechanism.
    #
    def get_masters(self):
        return [window for (title, window) in self._masters]
    
    #
    # Returns all the masters' ids.
    #
    def get_masters_by_id(self):
        return [window.id for (title, window) in self._masters]
    
    #
    # Returns the number of masters currently allowed.
    #
    def get_master_count(self):
        return self._master_count
    
    #
    # Returns all slaves currently in the storage.
    #
    def get_slaves(self):
        return [window for (title, window) in self._slaves]
    
    #
    # Returns all slaves' ids.
    #
    def get_slaves_by_id(self):
        return [window.id for (title, window) in self._slaves]
    
    #
    # Tells the tiling storage to allow for one more master.
    # There is no theoretical limit.
    #
    # Note: It is *possible* that your tiling algorithm will
    # want to restrict the number of masters. (See XMonad's
    # circle tiling algorithm.)
    #
    # Remember: This is not the number of masters, but rather
    # the number of *POSSIBLE* masters.
    #
    def inc_master_count(self):
        self._master_count += 1
        
    #
    # Removes a window from the storage.
    #
    def remove(self, window):
        if window.id in self.get_masters_by_id():
            self._remove_master(window)
        if window.id in self.get_slaves_by_id():
            self._remove_slave(window)
            
    #
    # Will sort the entire storage. This is currently
    # only called when you need to reload your storage.
    # It is not recommended to do so elsewhere, as it might
    # introduce inconsistencies between the storage and
    # the screen.
    #
    def sort(self):
        self._masters.sort()
        self._slaves.sort()
        
    #
    # Will switch any two windows. This preseves the
    # representation of storage as it pertains to the
    # screen's physical appearence.
    #
    def switch(self, win1, win2):
        for i in range(len(self._masters)):
            win = self._masters[i][1]
            if win1.id == win.id:
                self._masters[i] = (win2.title.lower(), win2)
            if win2.id == win.id:
                self._masters[i] = (win1.title.lower(), win1)
        
        for i in range(len(self._slaves)):
            win = self._slaves[i][1]
            if win1.id == win.id:
                self._slaves[i] = (win2.title.lower(), win2)
            if win2.id == win.id:
                self._slaves[i] = (win1.title.lower(), win1)
                
    #
    # This will try to promote a given window (has to be
    # a slave) to master status. Useful for reloading the
    # storage. (There could be slaves and extra room for
    # masters.)
    #
    def try_to_promote(self, window):
        if len(self._masters) < self.get_master_count() and window.id in self.get_slaves_by_id():
            self._remove_slave(window)
            self._add_master(window)
    
    
    #------------------------------------------------------------------------------
    # PRIVATE HELPER (INSTANCE) METHODS
    #------------------------------------------------------------------------------
    
    #
    # Explicitly adds a master to the storage. Please do not
    # call this directly, as it doesn't sync with the master
    # count.
    #
    def _add_master(self, window):
        self._masters.append(
                            (window.title.lower(), window)
                            )
        
    #
    # Explicitly adds a slave to the storage.
    #
    def _add_slave(self, window):
        self._slaves.append(
                           (window.title.lower(), window)
                           )
    
    #
    # Explicitly adds a master to the storage. Please do not
    # call this directly, as it doesn't sync with the master
    # count.
    #
    def _add_bottom_master(self, window):
        self._masters.append(
                            (window.title.lower(), window)
                            )
        
    #
    # Explicitly adds a slave to the storage.
    #
    def _add_bottom_slave(self, window):
        self._slaves.append(
                           (window.title.lower(), window)
                           )
        
    #
    # Explicitly adds a master to the storage. Please do not
    # call this directly, as it doesn't sync with the master
    # count.
    #
    def _add_top_master(self, window):
        self._masters.insert(0,
                            (window.title.lower(), window)
                            )
        
    #
    # Explicitly adds a slave to the storage.
    #
    def _add_top_slave(self, window):
        self._slaves.insert(0,
                           (window.title.lower(), window)
                           )
    
    #
    # Explicitly removes a master from the storage.
    #        
    def _remove_master(self, window):
        for i in range(len(self._masters)):
            win = self._masters[i][1]
            if window.id == win.id:
                del self._masters[i]
                break
     
    #
    # Explicitly removes a slave from the storage.
    #       
    def _remove_slave(self, window):
        for i in range(len(self._slaves)):
            win = self._slaves[i][1]
            if window.id == win.id:
                del self._slaves[i]
                break
    
    #
    # A nice output of the current storage. Useful for
    # debugging.
    #
    def __str__(self):
        ret = 'Master(s):\n'
        for master in self.get_masters():
            ret += '\t%s - %s\n' % (master.title, master.id)
        
        ret += 'Slave(s):\n'
        for slave in self.get_slaves():
            ret += '\t%s - %s\n' % (slave.title, slave.id)
            
        return ret


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Horizontal.py

The opposite to the Vertical layout. I like to use this on monitors with
1280x1024 (or 1920x1200) resolution, and Vertical on monitors with less
height to them (1280x720 or 1280x800).

Remember that this class inherits _cycle, help_find_next, and help_find_previous
from TileDefault.

Note that if you're creating your own tiling algorithm, you *MUST* add
    CLASS = CLASS_NAME
at the bottom of your class definition file (see the bottom of this file
for an example). This is so tiling algorithms can be dynamically loaded.
"""

from PyTyle.Tilers.TileDefault import TileDefault

class Horizontal (TileDefault):
    #
    # The core tiling algorithm. Every core tiling algorithm should start with
    # grabbing the current screen's workarea and factoring that into your
    # calculations. Feel free to follow my approach to tiling algorithms, or
    # come up with something else.
    #
    # Note: As I've been going through the source code writing these comments,
    # I've been thinking about generalizing tiling algorithms even more. So this
    # approach could change a little in the future. (Although I imagine the class
    # hierarchy will be staying the same, I like it.)
    #
    def _tile(self):
        x, y, width, height = self.screen.get_workarea()

        # set some vars...
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()

        masterWidth = width if not masters else (width / len(masters))
        masterHeight = height if not slaves else int(height * self.state.get('height_factor'))
        masterY = y
        masterX = x

        slaveWidth = width if not slaves else (width / len(slaves))
        slaveHeight = height if not masters else height - masterHeight
        slaveY = y if not masters else (y + masterHeight)
        slaveX = x

        # resize the master windows
        for master in masters:
            self.help_resize(master, masterX, masterY, masterWidth, masterHeight, self.state.get('margin'))
            masterX += masterWidth

        # now resize the rest... keep track of heights/positioning
        for slave in slaves:
            self.help_resize(slave, slaveX, slaveY, slaveWidth, slaveHeight, self.state.get('margin'))
            slaveX += slaveWidth

    #
    # Increases the height of all master windows. Don't forget to decrease
    # the height of all slave windows. Won't do anything if there are either
    # no masters or no slaves.
    #
    def _master_increase(self, factor = 0.05):
        x, y, width, height = self.screen.get_workarea()

        slaves = self.storage.get_slaves()
        masters = self.storage.get_masters()

        # Stop if neither of either... haha
        if not slaves or not masters:
            return

        # first calculate pixels...
        pixels = int(((self.state.get('height_factor') + factor) * height) - (self.state.get('height_factor') * height))
        self.state.set('height_factor', self.state.get('height_factor') + factor)

        for slave in slaves:
            slave.resize(slave.x, slave.y + pixels, slave.width, slave.height - pixels)
        for master in masters:
            master.resize(master.x, master.y, master.width, master.height + pixels)

    #
    # Decreases the height of all master windows. Don't forget to increase
    # the height of all slave windows. Won't do anything if there are either
    # no masters or no slaves.
    #
    def _master_decrease(self, factor = 0.05):
        x, y, width, height = self.screen.get_workarea()

        slaves = self.storage.get_slaves()
        masters = self.storage.get_masters()

        # Stop if neither of either... haha
        if not slaves or not masters:
            return

        # first calculate pixels...
        pixels = int((self.state.get('height_factor') * height) - ((self.state.get('height_factor') - factor) * height))
        self.state.set('height_factor', self.state.get('height_factor') - factor)

        for slave in slaves:
            slave.resize(slave.x, slave.y - pixels, slave.width, slave.height + pixels)
        for master in masters:
            master.resize(master.x, master.y, master.width, master.height - pixels)

# You must have this line's equivalent for your tiling algorithm!
# This makes it possible to dynamically load tiling algorithms.
# (So that you may simply drop them into the Tilers directory,
# and add their name to the configuration- vini, vidi, vicci!)
CLASS = Horizontal


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Desktop.py

A simple class to manage all desktops. It also includes a static
initialization method that will load all desktops reported by the
window manager. Each physical screen is initialized for each desktop.
This is also where tilers are initially set to screens.
"""

from PyTyle.Config import Config
from PyTyle.State import State
from PyTyle.Probe import PROBE

from PyTyle.Viewport import Viewport

class Desktop:
    #------------------------------------------------------------------------------
    # STATIC METHODS
    #------------------------------------------------------------------------------
    
    #
    # This is what really kicks it all off. It queries the window manager for all
    # available desktops, and instantiates them. When a desktop is initialized,
    # it will also load each of its screens. (Essentially, if xinerama reports two
    # available screens, then each screen will be attached to every desktop.)
    # Finally, this method also takes care of attaching a tiler to each screen,
    # which comes from the configuration file.
    #   
    @staticmethod
    def load_desktops():
        # initialize all desktops and their associated windows
        for desk in PROBE.get_desktops().values():
            desktop = Desktop(desk)
            for viewport in desktop.viewports.values():
                for screen in viewport.screens.values():
                    desk_or_view = desktop.id
                    if PROBE.is_compiz():
                        desk_or_view = viewport.id
                        
                    screen.set_tiler(Config.tilers(Config.tiling(screen.id, desk_or_view)))
                    
    #
    # Simply refreshes the desktop information. Used mainly when the workarea
    # changes to accomodate docks/panels.
    #
    @staticmethod
    def refresh_desktops():
        for desk in PROBE.get_desktops().values():
            if desk['id'] in State.get_desktops():
                desktop = State.get_desktops()[desk['id']]
                desktop.update_attributes(desk)
                for viewport in desktop.viewports.values():
                    for screen in viewport.screens.values():
                        screen.needs_tiling()
                    
    #
    # Same as "load_desktops" except we're just refreshing their information.
    #   
    @staticmethod
    def reload_desktops():
        # initialize all desktops and their associated windows
        for desktop in State.get_desktops().values():
            for viewport in desktop.viewports.values():
                for screen in viewport.screens.values():
                    desk_or_view = desktop.id
                    if PROBE.is_compiz():
                        desk_or_view = viewport.id
                        
                    screen.set_tiler(Config.tilers(Config.tiling(screen.id, desk_or_view)))
                    screen.needs_tiling()
                    
    
    #------------------------------------------------------------------------------
    # CONSTRUCTOR AND DESKTOP RELATED ATTRIBUTES/METHODS
    #------------------------------------------------------------------------------ 
                    
    #
    # The desktop constructor takes a dict of attributes fetched from X. It also
    # adds itself to the current State and loads all of its screens.
    #
    def __init__(self, attrs):
        self.update_attributes(attrs)
        self._VIEWPORT = None
        self.viewports = {}
        State.add_desktop(self)
        self.load_viewports()
        
    #
    # Probes X for all available viewports. For every desktop, an instance
    # of each viewport is newly created. (So the total number of "screens" 
    # in PyTyle is # of physical screens * viewports * desktops.)
    #
    def load_viewports(self):
        viewports = PROBE.get_viewports()
        for viewport in viewports:
            obj = Viewport(self, viewport)
            self.viewports[viewport['id']] = obj
            
    #
    # Simply updates all the desktop attributes. Currently only used in the
    # constructor. 
    #
    def update_attributes(self, attrs):
        self.id = attrs['id']
        self.resx = attrs['resx']
        self.resy = attrs['resy']
        self.x = attrs['x']
        self.y = attrs['y']
        self.width = attrs['width']
        self.height = attrs['height']
        self.name = attrs['name']
                
    #
    # Useful debugging string representation of the desktop. It prints
    # information about the current desktop, including each of its screens
    # and each window on each screen. See the string representations of
    # screen and window.
    #
    def __str__(self):
        retval = self.name + ' - [ID: ' + str(self.id) + ', RES: ' + str(self.resx) + 'x' + str(self.resy) + ', X: ' + str(self.x) + ', Y: ' + str(self.y) + ', WIDTH: ' + str(self.width) + ', HEIGHT: ' + str(self.height) + ']\n'
        
        for viewport in self.viewports.values():
            retval += '\t' + str(viewport) + '\n'
        
            for screen in viewport.screens.values():
                retval += '\t\t' + str(screen) + '\n'
                for window in screen.windows.values():
                    retval += '\t\t\t' + str(window) + '\n'
                retval += '\n'
            retval += '\n'
            
        return retval


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Window.py

Serves as a layer between tiling and interfacing with the Probe. It also
provides the function to load a window into PyTyle.

Windows are actually kept in three different data structures: (I should
probably change this so that Screen and TileStorage simply point to
windows in State?)
    State
        The State class keeps track of all windows currently known by
        PyTyle. This includes hidden windows, but doesn't include things
        like popups.
    Screen
        Each Screen keeps a record of which windows are on its screen.
        This will include hidden windows.
    TileStorage
        Each screen has its own tiling algorithm, and each tiling
        algorithm has its own tiling storage. In this storage, all tiling
        windows are kept. This will *not* include hidden windows.
"""

from PyTyle.Config import Config
from PyTyle.State import State
from PyTyle.Probe import PROBE

class Window:
    #------------------------------------------------------------------------------
    # STATIC METHODS
    #------------------------------------------------------------------------------

    #
    # This method uses the State to scan for any new windows (not currently in
    # the WINDOW state dict), and loads them into PyTyle. This is called at
    # program start up, and also when a new window has been created.
    #
    @staticmethod
    def load_new_windows():
        wins = State.scan_new_windows()
        for win in wins:
            Window.load_window(win)

    #
    # This loads a new window into PyTyle. It instantiates an object of this
    # class, and tells the window's screen that it needs to be retiled. It
    # will also activate itself if it's the active window. (This is
    # necessary because activating a window also updates the current
    # desktop and screen.)
    #
    # Note: This method also has logic to *skip* windows. Namely, it will
    # check to make sure that the window is in one of our stored desktops.
    # Also, it will make sure that it isn't a popup- otherwise it simply
    # won't be tiled.
    #
    @staticmethod
    def load_window(window_id):
        attrs = PROBE.get_window_by_id(window_id)
        if not attrs['popup'] and attrs['desktop'] in State.get_desktops():
            for viewport in State.get_desktops()[attrs['desktop']].viewports.values():
                if viewport.is_on_viewport(attrs['x'], attrs['y']):
                    for screen in viewport.screens.values():
                        if screen.is_on_screen(attrs['x'], attrs['y']):
                            win = Window(screen, attrs)
                            if not win.filtered():
                                screen.add_window(win)
                                screen.needs_tiling()

                                if win.id == PROBE.get_active_window_id():
                                    win.activate()


    #------------------------------------------------------------------------------
    # CONSTRUCTOR AND WINDOW RELATED ATTRIBUTES/METHODS
    #------------------------------------------------------------------------------

    #
    # The window constructor needs a screen to attach itself to, and a dict
    # of window attributes fetched from X. It also saves its current position so
    # that it may be "untiled." This is also where we'll start listening to the
    # window: (X.FocusChangeMask | X.StructureNotifyMask | X.PropertyChangeMask)
    #
    def __init__(self, screen, attrs):
        self.update_attributes(attrs)
        self.x = 0
        self.y = 0
        self.origx = attrs['x']
        self.origy = attrs['y']
        self.origwidth = self.width
        self.origheight = self.height
        self.screen = screen

        if self.xobj:
            PROBE.window_listen(self.xobj)

    #
    # Activates this window. It sets the input focus and also updates the
    # current State. (The desktop, screen, and active window.)
    #
    def activate(self):
        # we don't know where we are...
        if not self.screen:
            return

        PROBE.window_activate(self.xobj)
        State.reload_active(self)

    #
    # Attempts to add decorations to the window. It's currently only working
    # in OpenBox.
    #
    def add_decorations(self):
        PROBE.window_add_decorations(self.xobj)
        self.static = True

    #
    # Closes the current window. There are some problems with this right
    # now- namely, if I have any number of thunar windows open, closing
    # one of them will close all of them. However, if I close it using
    # OpenBox instead of PyTyle, everything is fine and dandy.
    #
    # Note: The close method in PyTyle isn't even needed. PyTyle can handle
    # the closing of a window in any way- nothing special happens if PyTyle
    # closes it. I am including it currently simply for completeness sake.
    # It *could* be removed in the future. I'm not sure. It is the only key
    # binding that doesn't do anything extra/different than any standard
    # WM event.
    #
    def close(self):
        PROBE.window_close(self.xobj)

    #
    # Deletes the window from existence. It also makes sure to queue its screen
    # for tiling.
    #
    def delete(self):
        self.screen.delete_window(self)
        self.screen.needs_tiling()
        State.reload_active()

    #
    # Runs the filter from the configuration on the given window.
    #
    # Note: This will check the window name, and class. It
    # might be a bit haphazard, but has worked so far. (Namely,
    # gmrun and gimp.)
    #
    # Note 2: It used to check the window title. But I saw that
    # Firefox doesn't set its download window as a pop up, and so
    # I had to use "download" to make PyTyle ignore it. That could
    # easily catch more windows than we want if we searched the
    # titles.
    #
    # Note 3: VLC (and quite probably, some other programs) aren't
    # reporting a proper class. I need something else to search by.
    #
    def filtered(self):
        if self.winclass:
            for winfilter in Config.filter():
                if self.winclass[0].lower().find(winfilter.lower()) != -1 or self.winclass[1].lower().find(winfilter.lower()) != -1: # or window.title.lower().find(winfilter.lower()) != -1:
                    return True

        return False

    #
    # Tests to see if this window is still alive.
    #
    def lives(self):
        try:
            PROBE.get_window_geometry(self.xobj)
        except:
            return False
        return True

    #
    # Asks the window manager to maximize the window. It does not currently
    # resize to the max screen coordinates.
    #
    def maximize(self):
        PROBE.window_maximize(self.xobj)

    #
    # Refreshs the current window. It probes for new window information and
    # updates the attributes on every call. It will then check if it has
    # moved to another desktop or screen. If it has, then it will queue both
    # the old and new screens to retile. It will then also update the State
    # information with the new desktop, screen, and window- if it's active.
    # See also the comments in the method for a few more interesting caveats.
    #
    def refresh(self):
        oldscreen = self.screen
        oldviewport = oldscreen.viewport
        olddesk = oldviewport.desktop
        oldstate = self.hidden
        update = PROBE.get_window(self.xobj)

        # So this is a little bit weird- we're updating the window, but while
        # we care about it's new x,y (screen change?), we don't care about it's
        # new width and height. We're tiling, so we're in complete control of
        # width and height. (The key here is that x/y *completely* determines
        # which screen the window is on.) Therefore, we don't update them- but
        # why? It would seem harmless, except that some windows (like terminals,
        # text editors, etc) set width_inc/height_inc hints which standards
        # compliant window managers honor- like OpenBox. Therefore, the WM could
        # be resizing the width/height of a given window slightly differently than
        # what PyTyle thinks it's at. This causes the width/height to change slightly
        # on each window update, and has a cascading effect that mangles the
        # window's size. YUCK. (And these width_inc/height_inc hints seemingly
        # cannot be reset.)
        update['width'] = self.width
        update['height'] = self.height
        self.update_attributes(update)

        if olddesk.id != self.desktop or not oldviewport.is_on_viewport(update['x'], update['y']) or not oldscreen.is_on_screen(update['x'], update['y']):
            for viewport in State.get_desktops()[self.desktop].viewports.values():
                if viewport.is_on_viewport(update['x'], update['y']):
                    for screen in viewport.screens.values():
                        if screen.is_on_screen(update['x'], update['y']):
                            oldscreen.delete_window(self)
                            screen.add_window(self)
                            screen.needs_tiling()
                            oldscreen.needs_tiling()
                            self.screen = screen
        elif oldstate != self.hidden:
            self.screen.needs_tiling()

        # If it's the active window, then make sure PyTyle knows that. We don't
        # want to set input focus (well, it should already have it if X tells us
        # it's the active window) because it will generate another window change
        # event, and we end up in a positive feedback loop. Yuck.
        if self.id == PROBE.get_active_window_id():
            State.reload_active()

    #
    # Attempts to remove decorations on the window. It's currently only
    # working in OpenBox.
    #
    # Note: devilspie seems to try to get it to work with most other
    # window managers as well. OpenBox has a special atom for the
    # "_NET_WM_STATE" property, namely, "_OB_WM_STATE_UNDECORATED".
    # But other window managers seem to use "_MOTIF_WM_HINTS" to
    # add/remove decorations. However, I could not get such a thing
    # to work in Fluxbox (it's the only other window manager I tried).
    #
    # Either way, I don't recommend using it- especially not yet. With
    # the decorations gone, it's pretty difficult to easily see which
    # window is the active window. (I haven't investigated drawing custom
    # borders around windows yet.)
    #
    def remove_decorations(self):
        PROBE.window_remove_decorations(self.xobj)
        self.static = False

    #
    # Yet another weird thing to handle. Sometimes windows will set their
    # own gravity to static (ex., Goggles Music Manager and OpenOffice).
    # Essentially, this changes how the window is positioned- and it really
    # mucks things up. The current solution I have is to simply change its
    # gravity to NorthWest (default). This seems to make everything play
    # nice, but I am weary of unpredictable behavior from windows that want
    # their gravity to be static, but instead I remove it. So far, so good,
    # though.
    #
    def remove_static_property(self):
        PROBE.window_remove_static(self.xobj)

    #
    # A simple resize method. This also updates the window with the given
    # coordinates.
    #
    # Note: It is possible (and perhaps, likely), that the real window
    # coordinates will become out of sync with the coordinates that PyTyle
    # has. Namely, if you're tiling a screen, you can still currently move
    # a window around that screen while it's floating. Since the
    # desktop/screen didn't change, PyTyle doesn't really care about it,
    # so nothing changes. So far though, nothing bad seems to have come
    # from it. PyTyle simply moves the window back to where it thinks it
    # should be on the next tiling action.
    #
    def resize(self, x, y, width, height):
        if width < 1 or height < 1 or not self.screen.is_in_screen(x, y, width, height):
            return

        PROBE.window_resize(self.xobj, x, y, width, height)
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    #
    # Asks the window to restore the window. This does not use the window's
    # original x/y/width/height saved in the constructor, but instead asks
    # the window manager to do it. (It should really just be called
    # "unmaximize".)
    #
    def restore(self):
        PROBE.window_reset(self.xobj)

    #
    # An on the fly way of saving the window's position. When this method is
    # called, it resets its original x/y/width/height, so that when it's
    # untiled, it will revert to this new position. Mainly useful for when
    # the window switches screens? (I'm not sure about that- if a window
    # goes to another screen on the same desktop, should it be untiled back
    # to its original screen, or an arbitrary position on its new screen?)
    #
    def save_geometry(self):
        geom = PROBE.get_window_geometry(self.xobj)
        self.origx = geom['x']
        self.origy = geom['y']
        self.origwidth = geom['width']
        self.origheight = geom['height']

    #
    # Lowers a window. Only use this if you're changing the stacking order.
    #
    def stack_lower(self):
        PROBE.window_stackbelow(self.xobj)

    #
    # Raises a window. Only use this if you're changing the stacking order.
    #
    def stack_raise(self):
        PROBE.window_stackabove(self.xobj)

    #
    # Simply updates all the window attributes.
    #
    def update_attributes(self, attrs):
        self.id = attrs['id']
#        self.x = attrs['x']
#        self.y = attrs['y']
        self.width = attrs['width']
        self.height = attrs['height']
        self.d_left = attrs['d_left']
        self.d_right = attrs['d_right']
        self.d_top = attrs['d_top']
        self.d_bottom = attrs['d_bottom']
        self.desktop = attrs['desktop']
        self.title = attrs['title']
        self.winclass = attrs['class']
        self.static = attrs['static']
        self.hidden = attrs['hidden']
        self.xobj = attrs['xobj'] if 'xobj' in attrs else None

    #
    # A simple string representation of the window. Useful for some debugging
    # purposes. Also see the string representations of desktop and screen.
    #
    def __str__(self):
        return self.title + ' - [ID: ' + str(self.id) + ', X: ' + str(self.x) + ', Y: ' + str(self.y) + ', WIDTH: ' + str(self.width) + ', HEIGHT: ' + str(self.height) + ', DESKTOP: ' + str(self.screen.viewport.desktop.id) + ', VIEWPORT: ' + str(self.screen.viewport.id) + ', SCREEN: ' + str(self.screen.id) + ']'



#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
TileDefault.py

TileDefault, in and of itself, cannot be used as a tiling algorithm. It
represents the methods that both the Vertical and Horizontal layouts have
in common. Both Vertical and Horizontal are subclasses of TileDefault.
"""

from PyTyle.Tile import Tile

class TileDefault (Tile):
    #------------------------------------------------------------------------------
    # OVERLOADED INSTANCE METHODS
    #------------------------------------------------------------------------------
    
    #
    # The common algorithm to cycle through all slaves for both the Horizontal
    # and Vertical layouts. Note that if there is more than one master, the first
    # master window will be swapped out. Also note that this will do nothing if
    # there are either no slaves or no masters. There will be nothing to cycle.
    #
    def _cycle(self):
        slaves = self.storage.get_slaves()
        masters = self.storage.get_masters()
        
        # Stop if neither of either... haha
        if not slaves or not masters:
            return
        
        # Might be a little high...
        if self.cycleIndex >= len(slaves):
            self.cycleIndex = len(slaves) - 1
        
        # use the index first to fetch the window to cycle
        self.help_switch(masters[0], slaves[self.cycleIndex])
        self.storage.get_masters()[0].activate()
        
        self.cycleIndex = self.cycleIndex + 1
        if self.cycleIndex == len(slaves): self.cycleIndex = 0
        
        
    #------------------------------------------------------------------------------
    # OVERLOADED PRIVATE HELPER METHODS
    #------------------------------------------------------------------------------ 
        
    #
    # Finds the next window. This is used both by the win_next and switch_next
    # tiling actions. It's a bit hairy, so I've commented the code as we go.
    #
    def help_find_next(self):
        # First retrieve all masters and all slaves... And also
        # pack them together.
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        all = masters + slaves
        
        # If there are masters, then we need to check if the currently
        # active window is the first master. If so, then we need to
        # focus on the first slave, or if there are no slaves, go to
        # the last master.
        if masters and self.screen.get_active().id == masters[0].id:
            if not slaves:
                return masters[-1]
            else:
                return slaves[0]
            
        # If there are slaves, then check if the active window is the
        # last slave. If it is, then focus on the last master, or if
        # there are no masters, then focus on the first slave.
        elif slaves and self.screen.get_active().id == slaves[-1].id:
            if not masters:
                return slaves[0]
            else:
                return masters[-1]
            
        # Now that our edge cases are satisfied, we simply find where
        # we are, and iterate to find the next window. (Same for masters
        # and slaves.)
        elif slaves and self.screen.get_active().id in [win.id for win in slaves]:
            for i in range(len(slaves) - 1):
                if self.screen.get_active().id == slaves[i].id:
                    return slaves[(i + 1)]
        elif masters:
            for i in range(1, len(masters)):
                if self.screen.get_active().id == masters[i].id:
                    return masters[(i - 1)]
        
    #
    # See help_find_next above. Also see the comments in the code,
    # as help_find_previous is basically the same thing- we're just
    # going in reverse.
    #
    def help_find_previous(self):
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        all = masters + slaves
        
        if masters and self.screen.get_active().id == masters[-1].id:
            if not slaves:
                return masters[0]
            else:
                return slaves[-1]
        elif slaves and self.screen.get_active().id == slaves[0].id:
            if not masters:
                return slaves[-1]
            else:
                return masters[0]
        elif masters and self.screen.get_active().id in [win.id for win in masters]:
            for i in range(len(masters) - 1):
                if self.screen.get_active().id == masters[i].id:
                    return masters[(i + 1)]
        elif slaves:
            for i in range(1, len(slaves)):
                if self.screen.get_active().id == slaves[i].id:
                    return slaves[(i - 1)]


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Tile.py

Tile is the guts of the tiling algorithm. It actually does *most* of the heavy
lifting for you, as many of the tiling actions (i.e., the key bindings in the
configuration) aren't really relevant to the physical display of the windows.
(However, the class hierarchy is such that you could make them relevant, if you
wish.)

In fact, to create your own customized tiling algorithm, you need only overload
the following methods: _tile, _cycle, _master_increase, _master_decrease,
help_find_next, and help_find_previous. By default, Vertical and Horizontal are
so similar, that currently, _cycle, help_find_next, and help_find_previous are
overloaded once in TileDefault (which cannot be used as a tiling algorithm by
itself), and _tile, _master_increase, and _master_decrease are each overloaded
accordingly in two of TileDefault's subclasses: Horizontal and Vertical.

To create your own tiling algorithm, read the comments in this file, along with
Tilers/TileDefault.py, Tilers/Horizontal.py, and Tilers/Vertical.py.
"""

import sys, os, time

from PyTyle.Config import Config
from PyTyle.State import State
from PyTyle.Debug import DEBUG
import traceback

from PyTyle.TileStorage import TileStorage
from PyTyle.TileState import TileState

class Tile:
    #------------------------------------------------------------------------------
    # STATIC METHODS (DISPATCHER RELATED)
    #------------------------------------------------------------------------------

    #
    # Initiate the dispatching routing.
    #
    # Has two contexts: Called with an action and called without.
    #
    # *Called with an action*
    # Implies that PyTyle is forcing the issue and wants something done. This also
    # means that we are here _not_ because of a key press. However, we still check
    # to see if tiling is enabled for this screen. If it isn't, we die.
    #
    # *Called without an action*
    # Implies that we have initiated the dispatch routine with a key press.
    # Therefore, we should check the currently pressed key, and follow the key
    # binding to run its respective method. Make sure to fail if tiling isn't
    # enabled and we aren't calling tile. (Essentially, pressing the tile key
    # binding is the only way to enable tiling.)
    #
    @staticmethod
    def dispatch(tiler, action=None, keycode=None, masks=None):
        if not action and keycode and masks:
            if keycode not in State.get_dispatcher():
                print >> sys.stderr, "Keycode %s is not bound" % keycode
                return

            # Now we need to determine which masks were used...
            if masks in State.get_dispatcher()[keycode]:
                action = State.get_dispatcher()[keycode][masks]
            else:
                print >> sys.stderr, "Keycode %s and keymask %d are not bound" % (keycode, masks)
                return

            if not tiler.screen.is_tiling() and action.find('tile.') == -1:
                return
        elif action:
            # We can only initiate tiling through keycodes...
            if not tiler.screen.is_tiling():
                return

        # Turn the action into a method...
        if action.find('tile.') != -1:
            layout = action[(action.find('.') + 1):]
            if layout != 'default' and layout in Config.TILERS:
                tiler.screen.set_tiler(Config.tilers(layout))
                tiler = tiler.screen.get_tiler()
                tiler._reset()
            action = Tile.tile
        else:
            action = eval('Tile.' + action)

        action(tiler)


    #------------------------------------------------------------------------------
    # CONSTRUCTOR AND GENERIC TILING METHODS
    #------------------------------------------------------------------------------

    #
    # Constructor simply instantiates the TileStorage class and passes it the window filter.
    # Every tiling instance must be attached to a screen.
    #
    # Also, in general, the following tiling methods are fairly generic (any method beginning
    # with an underscore). However, the class hierarchy is setup in such a way that does not
    # preclude their customization. Simply overload whichever method that needs customizing
    # in your tiling class. (_tile, _cycle, _master_increase, _master_decrease should be
    # sufficient here. Along with the helper methods help_find_next and help_find_previous.)
    # We also initialize this tiler's "state"- this will automatically save certain things for
    # us, like the sizes of panes.
    #
    def __init__(self, screen):
        self.screen = screen
        self.storage = TileStorage()
        self.cycleIndex = 0
        self.state = TileState(self)

    #
    # The core of the tiling algorithm. This will be called whenever PyTyle senses
    # that a screen needs to be re-tiled. It is essentially the bread and butter of
    # your tiling algorithm. It is responsible for placing *all* masters and slaves
    # in the TileStorage on the screen (unless you desire other behavior, like a
    # limited number of windows, but the behavior could be unpredictable). See the
    # respective tile methods in the algorithms shipped with this release.
    #
    def _tile(self):
        pass

    #
    # There is probably no need to overload this one. It simply iterates over all
    # stored windows and resizes them back to their original geometry. Geometry is
    # mainly saved when a window is first created and maybe when it switches screens.
    #
    def _untile(self):
        # just resize all the windows back to their original x/y/width/height
        for window in self.storage.get_all():
            if Config.misc('original_decor'):
                window.add_decorations()
            else:
                window.remove_decorations()

            window.resize(window.origx, window.origy, window.origwidth, window.origheight)

    #
    # Tells PyTyle to reload the configuration file.
    #
    def _reload(self):
        State.do_reload()

    #
    # Does a hard reset of the current screen. It empties the tiling storage, reloads
    # the current screen (probes for all windows), and adds the screen to the tiling
    # queue. Hopefully shouldn't have to be used, but is worth trying before
    # restarting PyTyle.
    #
    def _reset(self):
        self.storage = TileStorage()
        self.state.reset()
        self.cycleIndex = 0
        self.screen.needs_tiling()

    #
    # Responsible for cycling all slaves through the master slot. If there are more
    # than one master, then simply use the first master slot. If there are no masters
    # then do nothing. The cycleIndex should be incremented (or decremented, depending
    # upon your algorithm) after the cycle, and you also need to reset it to 0 when it
    # reaches the end. You should use help_find_next and help_find_previous here.
    #
    def _cycle(self):
        pass

    #
    # Simply puts focus on the given screen. The screen is responsible for providing
    # the currently active window. If we don't have one, then don't do anything. Also,
    # immediately quit if the screen is the same as the current tiler screen.
    #
    def _screen_focus(self, screen_num):
        # stop if current window...
        if self.screen.id == screen_num:
            return

        for screen in self.screen.viewport.screens.values():
            if screen_num == screen.id:
                if not screen.get_active():
                    return
                else:
                    screen.get_active().activate()

    #
    # Moves the active window to the given screen. Immediately quit if the screen is
    # the same as the current tiler screen.
    #
    # This method is a little tricky. Once we find our desired screen, we need to
    # update the storage in each screen (that is, the storage for the old screen
    # and the new screen- and this storage counts for both the tiler instance
    # and the screen instance). While we are only here if the current screen is in
    # tiling mode, we do however need to check if the screen we're moving the window
    # to is in tiling mode. If so, add the screen to the tiling queue (the current screen
    # also has to be added). Thus, if the screen we're moving it to is in tiling mode
    # then that screen is responsible for its placement. If not, arbitrarily assign
    # the window to the upper left coordinates of the screen. (Could be bad.) We
    # must then update the screen object of the window we moved, and activate the *old*
    # screen. (Incidentally, I did it this way because that is the default behavior
    # of XMonad.)
    #
    def _screen_put(self, screen_num):
        # stop if current screen...
        if self.screen.id == screen_num:
            return

        for screen in self.screen.viewport.screens.values():
            if screen_num == screen.id:
                add = self.screen.get_active()
                self.screen.delete_window(add)
                self.storage.remove(add)
                screen.add_window(add)
                screen.get_tiler().storage.add(add)

                if not screen.is_tiling():
                    add.resize(screen.x, screen.y, add.width, add.height)

                add.screen = screen
                self.screen.get_active().activate()

    #
    # Increases the area of the master pane. What this does is up to your tiling
    # algorithm. Pay special attention to the proper resizing of any other pane(s).
    #
    def _master_increase(self):
        pass

    #
    # Inverse of _master_increase.
    # _master_increase(_master_decrease(window placement)) = window placement
    # (Unless you set default pixel amounts of each method to be different... o_O)
    #
    def _master_decrease(self):
        pass

    #
    # Simply adds a master to the storage and submits the current screen into the
    # tiling queue. It is possible that your tiling algorithm might not want to
    # support more than one master (ex. XMonad's Circle layout), and in that case,
    # simply overload the method and leave it empty. (Make sure to do the same for
    # _remove_master or else you could be stuck!)
    #
    # Note: Both _add_master and _remove_master try to look at the currently active
    # window. If it is a slave or master, respectively, then that is the window that
    # will be added or removed. Otherwise, the first slave or master, respectively,
    # will be chosen arbitrarily.
    #
    def _add_master(self):
        # use active window if it's a slave
        slaves = self.storage.get_slaves()
        if self.screen.get_active().id in self.storage.get_slaves_by_id():
            self.storage.inc_master_count()
            self.storage.remove(self.screen.get_active())
            self.storage.add(self.screen.get_active())
        elif slaves:
            self.storage.inc_master_count()
            self.storage.remove(slaves[0])
            self.storage.add(slaves[0])
        else:
            return

        self.screen.needs_tiling()

    #
    # Inverse of _add_master.
    #
    # Note: If we somehow get a bigger master count than the number of windows (which
    # is okay and anticipated behavior), but we are trying to decrement, then decrement
    # the counter until we get to the current number of windows.
    #
    # Note: TileStorage will not let the number of masters go below 0 (0 masters is fine).
    #
    def _remove_master(self):
        # if we have too many windows, decrement master count until we're good...
        all = self.storage.get_all()
        while self.storage.get_master_count() > len(all):
            self.storage.dec_master_count()

        # make sure the current window is a master...
        masters = self.storage.get_masters()
        if self.screen.get_active().id in self.storage.get_masters_by_id():
            self.storage.dec_master_count()
            self.storage.remove(self.screen.get_active())
            self.storage.add(self.screen.get_active())
        elif masters:
            self.storage.dec_master_count()
            self.storage.remove(masters[0])
            self.storage.add(masters[0])
        else:
            return

        self.screen.needs_tiling()

    #
    # A very simple method to make the current window the master. If there are more
    # than one master, then use the first master. If there are no masters, then do
    # nothing.
    #
    def _make_active_master(self):
        if self.storage.get_masters():
            self.help_switch(self.storage.get_masters()[0], self.screen.get_active())

    #
    # Simply put the focus on the master window. If there are more than one master,
    # then use the first master. If there are no masters, then do nothing.
    #
    def _win_master(self):
        masters = self.storage.get_masters()

        if not masters:
            return

        masters[0].activate()

    #
    # Simply closes the current window.
    #
    # Note: We don't *really* need this method here. The window can be closed
    # in any way, and it will be detected by PyTyle. It's here mostly for
    # completeness.
    #
    def _win_close(self):
        self.screen.get_active().close()

    #
    # Focuses on the previous window.
    #
    def _win_previous(self):
        self.help_find_previous().activate()

    #
    # Focuses on the next window.
    #
    def _win_next(self):
        self.help_find_next().activate()

    #
    # Switches the current window with the previous window.
    #
    def _switch_previous(self):
        previous = self.help_find_previous()

        # only one window... bye
        if previous.id == self.screen.get_active().id:
            return

        self.help_switch(previous, self.screen.get_active())

    #
    # Switches the current window with the next window.
    #
    def _switch_next(self):
        next = self.help_find_next()

        # only one window... bye
        if next.id == self.screen.get_active().id:
            return

        self.help_switch(next, self.screen.get_active())

    #
    # Maximizes all windows managed by the tiler.
    #
    # Note: This sends a maximize request to the window manager. I'm not sure if this
    # is better than resizing the window to the full screen.
    #
    def _max_all(self):
        for window in self.storage.get_all():
            window.maximize()

    #
    # Restores all windows managed by the tiler.
    #
    def _restore_all(self):
        for window in self.storage.get_all():
            window.restore()

    #
    # A simple debugging tool. Only useful if PyTyle is running from a shell.
    # (Tentatively assigned the Alt-Q key binding.)
    #
    def _query(self):
        print State.get_wm_name()
        print self.screen.viewport.desktop
        print self.storage


    #------------------------------------------------------------------------------
    # PRIVATE HELPER METHODS
    #------------------------------------------------------------------------------

    #
    # Simply saves the position of all windows on this tiling screen.
    #
    def help_save(self):
        for window in self.screen.windows.values():
            window.save_geometry()

    #
    # Resizes the given window. Takes into account its decorations.
    #
    def help_resize(self, window, x, y, width, height, margin = 0):
        if margin > 0:
            x += margin
            y += margin
            width -= (2 * margin)
            height -= (2 * margin)

        if window.static:
            window.remove_static_property()

        if Config.misc('decorations'):
            if not Config.misc('original_decor'):
                window.add_decorations()

            window.resize(int(x), int(y), int(width - window.d_left - window.d_right), int(height - window.d_top - window.d_bottom))
        else:
            if Config.misc('original_decor'):
                window.remove_decorations()

            window.resize(int(x), int(y), int(width - 2), int(height - 2))

    #
    # Reloads the entire storage container underlying the tiling algorithm.
    # Unless you have really special needs, this should be sufficient. The
    # TilingStorage class handles master/slave balance for you according to
    # the masterCount. To use more or less masters, make calls to inc_master_count
    # and dec_master_count respectively.
    #
    # This method tries to update existing storage, although it doesn't assume
    # that there is currently any storage. Keep care to keep storage separate
    # from your tiling algorithm.
    #
    def help_reload(self):
        # delete first...
        for win in self.storage.get_all():
            if win.id not in self.screen.windows or win.hidden:
                self.storage.remove(win)

        # gobble up active window first in case we need a master...
        # and then just add away...
        masters = self.storage.get_masters_by_id()

        if self.screen.get_active() and len(masters) < self.storage.get_master_count() and self.screen.get_active().id in self.screen.windows and self.screen.get_active().id not in masters:
            self.storage.remove(self.screen.get_active())
            self.storage.add(self.screen.get_active())
            masters = self.storage.get_masters_by_id()

        all = self.storage.get_all_by_id()
        for window in self.screen.windows.values():
            if not window.id in all:
                self.storage.add_bottom(window)
            else:
                self.storage.try_to_promote(window)

    #
    # Simple method to switch two windows visually. It also takes care of
    # switching the windows in the storage container as well.
    #
    def help_switch(self, win1, win2):
        # same? weird...
        if win1.id == win2.id:
            return

        newpos = [win2.x, win2.y, win2.width, win2.height]
        win2.resize(win1.x, win1.y, win1.width, win1.height)
        win1.resize(newpos[0], newpos[1], newpos[2], newpos[3])

        self.storage.switch(win1, win2)

    #
    # The help_find_next and help_find_previous helper methods must be implemented
    # in a sub class if the functionality is to be used. Tiling
    # algorithms can sort windows differently such that the next/previous
    # windows cannot be accurately predicted generally. Unfortunately, these
    # little guys can be a little bit of a pain to right. (Edge cases, meh.)
    #
    def help_find_next(self):
        pass

    def help_find_previous(self):
        pass


    #------------------------------------------------------------------------------
    # DISPATCH
    #
    # The following methods are called from the dispatcher, and are directly
    # referenced from inside the .pytylerc configuration file. Most of these
    # methods simply call their member equivalents, but some (like tile/untile)
    # do some house keeping before hand.
    #
    # These should not be overloaded, but rather their equivalent member methods
    # preceded with an "_" should be overloaded to customize your tiling algorithm.
    #
    # You can get cursory information about these methods from the configuration
    # file, or you may peruse their comments above (in this class), and/or may
    # also take a look at the comments in Tilers/TileDefault.py,
    # Tilers/Vertical.py, and/or Tilers/Horizontal.py.
    #------------------------------------------------------------------------------

    def tile(self):
        # save state...
        if not self.screen.is_tiling():
            self.help_save()

        # If we haven't tiled and we're about to,
        # reload the storage...
        if not self.screen.is_tiled():
            self.help_reload()

        self.screen.enable_tiling()
        self._tile()
        self.screen.got_tiling()

    def untile(self):
        self._untile()
        self.screen.disable_tiling()

    def cycle_tiler(self):
        for i in range(len(Config.misc('tilers'))):
            if Config.misc('tilers')[i] is self.__class__.__name__:
                if (i + 1) == len(Config.misc('tilers')):
                    self.screen.set_tiler(Config.tilers(Config.misc('tilers')[0]))
                else:
                    self.screen.set_tiler(Config.tilers(Config.misc('tilers')[i + 1]))

        self._reset()

    def reload(self):
        self._reload()

    def reset(self):
        self._reset()

    def cycle(self):
        self._cycle()

    def screen0_focus(self):
        self._screen_focus(0)

    def screen1_focus(self):
        self._screen_focus(1)

    def screen2_focus(self):
        self._screen_focus(2)

    def screen0_put(self):
        self._screen_put(0)

    def screen1_put(self):
        self._screen_put(1)

    def screen2_put(self):
        self._screen_put(2)

    def master_increase(self):
        self._master_increase()

    def master_decrease(self):
        self._master_decrease()

    def add_master(self):
        self._add_master()

    def remove_master(self):
        self._remove_master()

    def make_active_master(self):
        self._make_active_master()

    def win_master(self):
        self._win_master()

    def win_close(self):
        self._win_close()

    def win_previous(self):
        self._win_previous()

    def win_next(self):
        self._win_next()

    def switch_previous(self):
        self._switch_previous()

    def switch_next(self):
        self._switch_next()

    def max_all(self):
        self._max_all()

    def restore_all(self):
        self._restore_all()

    def query(self):
        self._query()


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

import sys
from distutils import sysconfig
from distutils.core import setup

try:
    from Xlib.display import Display
    from Xlib import X, XK, Xatom, Xutil, protocol
    from Xlib.ext import xinerama
except:
    print ''
    print 'PyTyle requires the Python X Library'
    print 'See: http://python-xlib.sourceforge.net/'
    sys.exit(0)

setup(
      name = "pytyle",
      author = "Andrew Gallant",
      author_email = "andrew@pytyle.com",
      version = "0.7.5",
      license = "GPL",
      description = "A manual tiling manager for EWMH compliant window managers",
      long_description = "See README",
      url = "http://pytyle.com",
      platforms = 'POSIX',
      packages = ['PyTyle', 'PyTyle.Tilers'],
      data_files = [
                    (sysconfig.get_python_lib() + '/PyTyle',
                     ['./pytylerc', './INSTALL', './LICENSE', './README', './TODO', './CHANGELOG'])
                    ],
      scripts = ['pytyle','pytyle-client']
      )



#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
Cascade.py

Performs a full screen cascading layout of all the windows. It's important that
we're able to access the window decoration sizes.
"""

from PyTyle.Tilers.TileDefault import TileDefault

class Cascade (TileDefault):
    #------------------------------------------------------------------------------
    # OVERLOADED INSTANCE METHODS
    #------------------------------------------------------------------------------
        
    #
    # This Cascade layout will essentially do the following:
    #    1. The first window (or bottom) window takes up the full screen.
    #    2. Each subsequent window has its height shrinked by the height of
    #       the window decoration.
    #
    def _tile(self):
        x, y, width, height = self.screen.get_workarea()
        
        # set some vars...
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        
        # If we don't have any decorations... use hard coded value for now
        if masters and masters[0].d_top:
            decor = masters[0].d_top
        elif slaves and slaves[0].d_top:
            decor = slaves[0].d_top
        else:
            decor = self.state.get('decoration_height')
        
        push_over = self.state.get('push_over')
        push_width = push_over
        if self.state.get('horz_align') == 'right':
            push_over = -push_over
        
        masterWidth = (width * self.state.get('width_factor')) - (push_width * len(slaves))
        masterHeight = (height * self.state.get('height_factor')) - (decor * len(slaves))
        masterY = y + (decor * len(slaves))
        
        slaveWidth = width * self.state.get('width_factor')
        slaveHeight = height * self.state.get('height_factor')
        slaveY = y
        
        if self.state.get('horz_align') == 'right':
            masterX = x + (width - masterWidth) + (push_over * len(slaves))
            slaveX = x + (width - slaveWidth)
            push_over = 0
        else:
            masterX = x + (push_over * len(slaves))
            slaveX = x
        
        # now resize the rest... keep track of heights/positioning
        for slave in slaves:
            self.help_resize(slave, slaveX, slaveY, slaveWidth, slaveHeight)
            slaveY += decor
            slaveHeight -= decor
            slaveWidth -= push_width
            slaveX += push_over
            slave.stack_raise()
            
        # resize the master windows
        for master in masters:
            self.help_resize(master, masterX, masterY, masterWidth, masterHeight)
            master.stack_raise()
            
        # just in case...
        self.screen.get_active().stack_raise()
    
    #
    # Not changing much functionality here. Just overloading, inheriting, and
    # making sure the stacking order is kept.
    #
    def _cycle(self):
        TileDefault._cycle(self)
        self.help_reset_stack()
        
    def _make_active_master(self):
        TileDefault._make_active_master(self)
        self.help_reset_stack()
        
    def _win_master(self):
        TileDefault._win_master(self)
        self.help_reset_stack()

    def _win_previous(self):
        self.help_find_previous().activate()
        self.help_reset_stack()

    def _switch_previous(self):
        TileDefault._switch_previous(self)        
        self.help_reset_stack()  

    def _switch_next(self):
        TileDefault._switch_next(self)        
        self.help_reset_stack()
        
    #
    # We want to disable the following methods. They are of no use for this
    # layout.
    # 
    def _master_increase(self, pixels = 50):
        pass
    
    def _master_decrease(self, pixels = 50):
        pass
    
    def _add_master(self):
        pass
    
    def _remove_master(self):
        pass 
    
    
    #------------------------------------------------------------------------------
    # PRIVATE HELPER METHODS
    #------------------------------------------------------------------------------
    
    #
    # Same exact thing as Tile.help_reload, except we add to the top
    # of the window stack instead.
    #
    def help_reload(self):
        # delete first...
        for win in self.storage.get_all():
            if win.id not in self.screen.windows or win.hidden:
                self.storage.remove(win)
                
        # gobble up active window first in case we need a master...
        # and then just add away...
        masters = self.storage.get_masters_by_id()
        
        if self.screen.get_active() and len(masters) < self.storage.get_master_count() and self.screen.get_active().id in self.screen.windows and self.screen.get_active().id not in masters:
            self.storage.remove(self.screen.get_active())
            self.storage.add(self.screen.get_active())
            masters = self.storage.get_masters_by_id()                
        
        all = self.storage.get_all_by_id()
        for window in self.screen.windows.values():
            if not window.id in all:
                self.storage.add_top(window)
            else:
                self.storage.try_to_promote(window) 
     
    #
    # Resets the stacking order. I'm not sure if this is the best way to
    # do it, but it seems fairly quick. There's a small flicker, but that's
    # it. Maybe be more selective with its use?
    # 
    def help_reset_stack(self):
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        
        # now set the stacking order straight...
        for slave in slaves:
            slave.stack_raise()
        self.screen.get_active().stack_raise()
            
# You must have this line's equivalent for your tiling algorithm!
# This makes it possible to dynamically load tiling algorithms.
# (So that you may simply drop them into the Tilers directory,
# and add their name to the configuration- vini, vidi, vicci!)
CLASS = Cascade


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
TileState.py

Keeps a record of the state of your tiler. This is helpful for automatically
keeping window sizes consistent when new windows are added, removed, etc. 

TileState also interfaces with the LAYOUT config.
"""

from PyTyle.Config import Config

class TileState:
    #------------------------------------------------------------------------------
    # CONSTRUCTOR AND INSTANCE METHODS
    #------------------------------------------------------------------------------
    
    #
    # Just starts the state and tiler instance variables.
    #
    def __init__(self, tiler):
        self._state = {}
        self._tiler = tiler
        
    #
    # Retrieves a state item. First we check our current state.
    # If we don't have a record of it, then look in the layout
    # config (and copy this value to our current state). Otherwise
    # return nothing- we don't have a state yet.
    #
    def get(self, key):
        if key in self._state:
            return self._state[key]
        elif Config.layout(self._tiler, key) is not None:
            self.set(key, Config.layout(self._tiler, key))
            return Config.layout(self._tiler, key)
        return None
    
    #
    # Empties the current state. Remember, the state *starts* as
    # empty, so tiling algorithms need to make sure they can
    # handle bad values. (Or put it in the config.)
    #
    def reset(self):
        self._state = {}
    
    #
    # Sets a state item.
    #
    def set(self, key, value):
        self._state[key] = value


#===============================================================================
# PyTyle - A manual tiling manager
# Copyright (C) 2009  Andrew Gallant <andrew@pytyle.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===============================================================================

"""
HorizontalRows.py

This is something *like* the "Horizontal" layout, except instead of having
one row of slaves "beneath" the master, it can have multiple rows. By
default, it will be two windows per row (or, two columns). This can be
changed in the LAYOUT portion of the configuration file.
"""

from PyTyle.Tilers.TileDefault import TileDefault
import math

class HorizontalRows (TileDefault):
    #
    # Does almost the same thing as the Horizontal layout,
    # but is a bit more complex to account for multiple
    # rows.
    #
    def _tile(self):
        x, y, width, height = self.screen.get_workarea()

        # set some vars...
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        row_size = self.state.get('row_size')
        rows = int(math.ceil(float(len(slaves)) / float(row_size)))
        last_row_size = len(slaves) % row_size
        if not last_row_size: last_row_size = row_size

        masterWidth = width if not masters else (width / len(masters))
        masterHeight = height if not slaves else int(height * self.state.get('height_factor'))
        masterY = y
        masterX = x

        slaveWidth = width if not slaves else (width / row_size)
        slaveHeight = height if not slaves else ((height if not masters else height - masterHeight) / rows)
        slaveY = y if not masters else (y + masterHeight)
        slaveX = x

        # resize the master windows
        for master in masters:
            self.help_resize(master, masterX, masterY, masterWidth, masterHeight, self.state.get('margin'))
            masterX += masterWidth

        current_row = 1
        current_window = 1
        for slave in slaves:
            # last row!
            if current_row == rows:
                slaveWidth = width / last_row_size

            self.help_resize(slave, slaveX, slaveY, slaveWidth, slaveHeight, self.state.get('margin'))
            slaveX += slaveWidth

            if current_window % row_size == 0:
                current_row += 1
                current_window = 1
                slaveX = x
                slaveY += slaveHeight
            else:
                current_window += 1

    #
    # This one is actually pretty neat. First we need to take care
    # of our masters, which is pretty straight forward. Then we move
    # the first row of slaves up by the regular pixel amount. Then the
    # second row gets moved up slightly less (pixels - (pixels/rows))
    # to be exact. And so on and so forth. Each row gets a height
    # increase of pixels/rows.
    #
    # Note: We of course deal with integers here. In order to get around
    # what could be a nasty rounding problem, we modify the pixel
    # increase/decrease number right away to make sure it's a multiple
    # of the number of rows. (We go down instead of up.)
    #
    def _master_increase(self, factor = 0.05):
        x, y, width, height = self.screen.get_workarea()
        slaves = self.storage.get_slaves()
        masters = self.storage.get_masters()
        row_size = self.state.get('row_size')
        rows = int(math.ceil(float(len(slaves)) / float(row_size)))

        # Stop if neither of either... haha
        if not slaves or not masters:
            return

        # first calculate pixels...
        pixels = int(((self.state.get('height_factor') + factor) * height) - (self.state.get('height_factor') * height))
        self.state.set('height_factor', self.state.get('height_factor') + factor)

        # Make it easy... change pixels to next closest
        # multiple of the number of rows...
        pixels = pixels - (pixels % rows)
        slave_pixels = pixels / rows

        current_row = 1
        current_window = 1
        height_pixels = pixels
        for slave in slaves:
            slave.resize(slave.x, slave.y + height_pixels, slave.width, slave.height - slave_pixels)

            if current_window % row_size == 0:
                current_row += 1
                current_window = 1
                height_pixels -= slave_pixels
            else:
                current_window += 1
        for master in masters:
            master.resize(master.x, master.y, master.width, master.height + pixels)

    #
    # See master_increase.
    #
    def _master_decrease(self, factor = 0.05):
        x, y, width, height = self.screen.get_workarea()
        slaves = self.storage.get_slaves()
        masters = self.storage.get_masters()
        row_size = self.state.get('row_size')
        rows = int(math.ceil(float(len(slaves)) / float(row_size)))

        # Stop if neither of either... haha
        if not slaves or not masters:
            return

        # first calculate pixels...
        pixels = int((self.state.get('height_factor') * height) - ((self.state.get('height_factor') - factor) * height))
        self.state.set('height_factor', self.state.get('height_factor') - factor)

        # Make it easy... change pixels to next closest
        # multiple of the number of rows...
        pixels = pixels - (pixels % rows)
        slave_pixels = pixels / rows

        current_row = 1
        current_window = 1
        height_pixels = pixels
        for slave in slaves:
            slave.resize(slave.x, slave.y - height_pixels, slave.width, slave.height + slave_pixels)

            if current_window % row_size == 0:
                current_row += 1
                current_window = 1
                height_pixels -= slave_pixels
            else:
                current_window += 1
        for master in masters:
            master.resize(master.x, master.y, master.width, master.height - pixels)

    #------------------------------------------------------------------------------
    # OVERLOADED PRIVATE HELPER METHODS
    #------------------------------------------------------------------------------


    #
    # In HorizontalRows, the alignment of the masters is a bit weird. We would like
    # to have a consistent ordering style like so:
    # win --- win
    #       -
    #      -
    #     -
    # win --- win
    #       -
    #      -
    #     -
    # win --- win
    #
    # Note: See this method in Tile.TileDefault for additional comments.
    #
    def help_find_next(self):
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        all = masters + slaves

        if masters and self.screen.get_active().id == masters[-1].id:
            if not slaves:
                return masters[0]
            else:
                return slaves[0]
        elif slaves and self.screen.get_active().id == slaves[-1].id:
            if not masters:
                return slaves[0]
            else:
                return masters[0]
        elif slaves and self.screen.get_active().id in [win.id for win in slaves]:
            for i in range(len(slaves) - 1):
                if self.screen.get_active().id == slaves[i].id:
                    return slaves[(i + 1)]
        elif masters:
            for i in range(0, len(masters) - 1):
                if self.screen.get_active().id == masters[i].id:
                    return masters[(i + 1)]

    #
    # See help_find_next above. Also see the comments in the code,
    # as help_find_previous is basically the same thing- we're just
    # going in reverse.
    #
    def help_find_previous(self):
        masters = self.storage.get_masters()
        slaves = self.storage.get_slaves()
        all = masters + slaves

        if masters and self.screen.get_active().id == masters[0].id:
            if not slaves:
                return masters[-1]
            else:
                return slaves[-1]
        elif slaves and self.screen.get_active().id == slaves[0].id:
            if not masters:
                return slaves[-1]
            else:
                return masters[-1]
        elif masters and self.screen.get_active().id in [win.id for win in masters]:
            for i in range(1, len(masters)):
                if self.screen.get_active().id == masters[i].id:
                    return masters[(i - 1)]
        elif slaves:
            for i in range(1, len(slaves)):
                if self.screen.get_active().id == slaves[i].id:
                    return slaves[(i - 1)]

# You must have this line's equivalent for your tiling algorithm!
# This makes it possible to dynamically load tiling algorithms.
# (So that you may simply drop them into the Tilers directory,
# and add their name to the configuration- vini, vidi, vicci!)
CLASS = HorizontalRows